
from math import log

import pytest

import wn
from wn import similarity as sim
from wn.taxonomy import taxonomy_depth
from wn.ic import information_content as infocont


def get_synsets(w):
    return {
        'information': w.synset('test-en-0001-n'),
        'example': w.synset('test-en-0002-n'),
        'sample': w.synset('test-en-0004-n'),
        'random sample': w.synset('test-en-0005-n'),
        'random sample2': w.synset('test-en-0008-n'),
        'datum': w.synset('test-en-0006-n'),
        'exemplify': w.synset('test-en-0003-v'),
    }


# some fake information content; computed using:
#   words = ['example', 'example', 'sample', 'random sample', 'illustrate']
#   ic = compute(words, wn.Wordnet('test-en'), distribute_weight=False)

ic = {
    'n': {'test-en-0001-n': 5.0,  # information
          'test-en-0002-n': 5.0,  # example, illustration
          'test-en-0004-n': 3.0,  # sample
          'test-en-0005-n': 2.0,  # random sample
          'test-en-0008-n': 2.0,  # random sample 2
          'test-en-0006-n': 1.0,  # datum
          None: 6.0},
    'v': {'test-en-0003-v': 2.0,  # exemplify, illustrate
          'test-en-0007-v': 1.0,  # resignate
          None: 2.0},
    'a': {None: 1.0},
    'r': {None: 1.0}
}


@pytest.mark.usefixtures('mini_db')
def test_path():
    ss = get_synsets(wn.Wordnet('test-en'))
    assert sim.path(ss['information'], ss['information']) == 1/1
    assert sim.path(ss['information'], ss['example']) == 1/2
    assert sim.path(ss['information'], ss['sample']) == 1/3
    assert sim.path(ss['information'], ss['random sample']) == 1/4
    assert sim.path(ss['random sample'], ss['datum']) == 1/5
    assert sim.path(ss['random sample2'], ss['datum']) == 0
    assert sim.path(ss['random sample2'], ss['datum'], simulate_root=True) == 1/4
    assert sim.path(ss['random sample'], ss['random sample2'], simulate_root=True) == 1/6
    with pytest.raises(wn.Error):
        sim.path(ss['example'], ss['exemplify'])
    with pytest.raises(wn.Error):
        sim.wup(ss['example'], ss['exemplify'], simulate_root=True)


@pytest.mark.usefixtures('mini_db')
def test_wup():
    ss = get_synsets(wn.Wordnet('test-en'))
    assert sim.wup(ss['information'], ss['information']) == (2*1) / (0+0+2*1)
    assert sim.wup(ss['information'], ss['example']) == (2*1) / (0+1+2*1)
    assert sim.wup(ss['information'], ss['sample']) == (2*1) / (0+2+2*1)
    assert sim.wup(ss['information'], ss['random sample']) == (2*1) / (0+3+2*1)
    assert sim.wup(ss['random sample'], ss['datum']) == (2*1) / (3+1+2*1)
    with pytest.raises(wn.Error):
        assert sim.wup(ss['random sample2'], ss['datum'])
    assert (sim.wup(ss['random sample2'], ss['datum'], simulate_root=True)
            == (2*1) / (1+2+2*1))
    assert (sim.wup(ss['random sample'], ss['random sample2'], simulate_root=True)
            == (2*1) / (4+1+2*1))
    with pytest.raises(wn.Error):
        sim.wup(ss['example'], ss['exemplify'])
    with pytest.raises(wn.Error):
        sim.wup(ss['example'], ss['exemplify'], simulate_root=True)


@pytest.mark.usefixtures('mini_db')
def test_lch():
    w = wn.Wordnet('test-en')
    ss = get_synsets(w)
    d_n = taxonomy_depth(w, 'n')
    assert sim.lch(ss['information'], ss['information'], d_n) == -log((0+1) / (2*d_n))
    assert sim.lch(ss['information'], ss['example'], d_n) == -log((1+1) / (2*d_n))
    assert sim.lch(ss['information'], ss['sample'], d_n) == -log((2+1) / (2*d_n))
    assert sim.lch(ss['information'], ss['random sample'], d_n) == -log((3+1) / (2*d_n))
    assert sim.lch(ss['random sample'], ss['datum'], d_n) == -log((4+1) / (2*d_n))
    with pytest.raises(wn.Error):
        assert sim.lch(ss['random sample2'], ss['datum'], d_n)
    assert (sim.lch(ss['random sample2'], ss['datum'], d_n, simulate_root=True)
            == -log((3+1) / (2*d_n)))
    assert (sim.lch(ss['random sample'], ss['random sample2'], d_n, simulate_root=True)
            == -log((5+1) / (2*d_n)))
    with pytest.raises(wn.Error):
        sim.lch(ss['example'], ss['exemplify'], d_n)
    with pytest.raises(wn.Error):
        sim.lch(ss['example'], ss['exemplify'], d_n, simulate_root=True)


@pytest.mark.usefixtures('mini_db')
def test_res():
    w = wn.Wordnet('test-en')
    ss = get_synsets(w)
    assert (sim.res(ss['information'], ss['information'], ic)
            == infocont(ss['information'], ic))
    assert (sim.res(ss['information'], ss['example'], ic)
            == infocont(ss['information'], ic))
    assert (sim.res(ss['information'], ss['sample'], ic)
            == infocont(ss['information'], ic))
    assert (sim.res(ss['information'], ss['random sample'], ic)
            == infocont(ss['information'], ic))
    assert (sim.res(ss['random sample'], ss['datum'], ic)
            == infocont(ss['information'], ic))
    with pytest.raises(wn.Error):
        sim.res(ss['random sample2'], ss['datum'], ic)
    with pytest.raises(wn.Error):
        sim.res(ss['example'], ss['exemplify'], ic)


@pytest.mark.usefixtures('mini_db')
def test_jcn():
    w = wn.Wordnet('test-en')
    ss = get_synsets(w)
    info_ic = infocont(ss['information'], ic)
    assert (sim.jcn(ss['information'], ss['information'], ic)
            == float('inf'))
    assert (sim.jcn(ss['information'], ss['example'], ic)
            == float('inf'))
    assert (sim.jcn(ss['information'], ss['sample'], ic)
            == 1 / ((info_ic + infocont(ss['sample'], ic)) - 2 * info_ic))
    assert (sim.jcn(ss['information'], ss['random sample'], ic)
            == 1 / ((info_ic + infocont(ss['random sample'], ic)) - 2 * info_ic))
    assert (sim.jcn(ss['random sample'], ss['datum'], ic)
            == 1 / (
                (infocont(ss['random sample'], ic) + infocont(ss['datum'], ic))
                - 2 * info_ic))
    with pytest.raises(wn.Error):
        sim.jcn(ss['random sample2'], ss['datum'], ic)
    with pytest.raises(wn.Error):
        sim.jcn(ss['example'], ss['exemplify'], ic)


@pytest.mark.usefixtures('mini_db')
def test_lin():
    w = wn.Wordnet('test-en')
    ss = get_synsets(w)
    info_ic = infocont(ss['information'], ic)
    assert (sim.lin(ss['information'], ss['information'], ic)
            == 1.0)
    assert (sim.lin(ss['information'], ss['example'], ic)
            == 1.0)
    assert (sim.lin(ss['information'], ss['sample'], ic)
            == (2 * info_ic) / (info_ic + infocont(ss['sample'], ic)))
    assert (sim.lin(ss['information'], ss['random sample'], ic)
            == (2 * info_ic) / (info_ic + infocont(ss['random sample'], ic)))
    assert (sim.lin(ss['random sample'], ss['datum'], ic)
            == ((2 * info_ic)
                / (infocont(ss['random sample'], ic) + infocont(ss['datum'], ic))))
    with pytest.raises(wn.Error):
        sim.lin(ss['random sample2'], ss['datum'], ic)
    with pytest.raises(wn.Error):
        sim.lin(ss['example'], ss['exemplify'], ic)




from xml.etree import ElementTree as ET

from wn import lmf


def test_is_lmf(datadir):
    assert lmf.is_lmf(datadir / 'mini-lmf-1.0.xml')
    assert lmf.is_lmf(str(datadir / 'mini-lmf-1.0.xml'))
    assert not lmf.is_lmf(datadir / 'README.md')
    assert not lmf.is_lmf(datadir / 'missing.xml')
    assert lmf.is_lmf(datadir / 'mini-lmf-1.1.xml')


def test_load_1_0(mini_lmf_1_0):
    resource = lmf.load(mini_lmf_1_0)
    lexicons = resource['lexicons']
    assert len(lexicons) == 2
    lexicon = lexicons[0]

    assert lexicon['id'] == 'test-en'
    assert lexicon['label'] == 'Testing English WordNet'
    assert lexicon['language'] == 'en'
    assert lexicon['email'] == 'maintainer@example.com'
    assert lexicon['license'] == 'https://creativecommons.org/licenses/by/4.0/'
    assert lexicon['version'] == '1'
    assert lexicon['url'] == 'https://example.com/test-en'

    assert len(lexicon['entries']) == 9
    le = lexicon['entries'][0]
    assert le['id'] == 'test-en-information-n'

    assert le['lemma']['writtenForm'] == 'information'
    assert le['lemma']['partOfSpeech'] == 'n'
    assert le['lemma']['script'] == 'Latn'
    assert len(le['lemma']['tags']) == 1

    assert len(le.get('forms', [])) == 0

    assert len(le['senses']) == 1
    sense = le['senses'][0]
    assert sense['id'] == 'test-en-information-n-0001-01'
    assert sense['synset'] == 'test-en-0001-n'
    assert len(sense.get('relations', [])) == 0
    # assert sense['relations'][0]['target'] == 'test-en-exemplify-v-01023137-01'
    # assert sense['relations'][0]['type'] == 'derivation'

    assert len(lexicon.get('frames', [])) == 0  # frames are on lexical entry
    assert len(lexicon['entries'][6]['frames']) == 2
    frames = lexicon['entries'][6]['frames']
    assert frames[0]['subcategorizationFrame'] == 'Somebody ----s something'
    assert frames[0]['senses'] == ['test-en-illustrate-v-0004-01']

    assert len(lexicon['synsets']) == 8

    assert lexicons[1]['id'] == 'test-es'


def test_load_1_1(mini_lmf_1_1):
    resource = lmf.load(mini_lmf_1_1)
    lexicons = resource['lexicons']
    assert len(lexicons) == 2
    lexicon = lexicons[0]
    assert lexicon['id'] == 'test-ja'
    assert lexicon['version'] == '1'
    # assert lexicon.logo == 'logo.svg'
    assert lexicon.get('requires') == [{'id': 'test-en', 'version': '1'}]

    lexicon = lexicons[1]
    assert lexicon['id'] == 'test-en-ext'
    assert lexicon.get('extends') == {'id': 'test-en', 'version': '1'}


def test_dump(mini_lmf_1_0, mini_lmf_1_1, tmp_path):
    tmpdir = tmp_path / 'test_dump'
    tmpdir.mkdir()
    tmppath = tmpdir / 'mini_lmf_dump.xml'

    def assert_xml_equal(mini_lmf, dump_lmf):
        if hasattr(ET, 'canonicalize'):  # available from Python 3.8
            orig = ET.canonicalize(from_file=mini_lmf, strip_text=True)
            temp = ET.canonicalize(from_file=dump_lmf, strip_text=True)
            # additional transformation to help with debugging
            orig = orig.replace('<', '\n<')
            temp = temp.replace('<', '\n<')
            assert orig == temp

    lmf.dump(lmf.load(mini_lmf_1_0), tmppath)
    assert_xml_equal(mini_lmf_1_0, tmppath)

    lmf.dump(lmf.load(mini_lmf_1_1), tmppath)
    assert_xml_equal(mini_lmf_1_1, tmppath)



"""
Constants and literals used in wordnets.
"""


SENSE_RELATIONS = frozenset([
    'antonym',
    'also',
    'participle',
    'pertainym',
    'derivation',
    'domain_topic',
    'has_domain_topic',
    'domain_region',
    'has_domain_region',
    'exemplifies',
    'is_exemplified_by',
    'similar',
    'other',
    'feminine',
    'has_feminine',
    'masculine',
    'has_masculine',
    'young',
    'has_young',
    'diminutive',
    'has_diminutive',
    'augmentative',
    'has_augmentative',
    'anto_gradable',
    'anto_simple',
    'anto_converse',
    'simple_aspect_ip',
    'secondary_aspect_ip',
    'simple_aspect_pi',
    'secondary_aspect_pi',
])

SENSE_SYNSET_RELATIONS = frozenset([
    'other',
    'domain_topic',
    'domain_region',
    'exemplifies',
])

SYNSET_RELATIONS = frozenset([
    'agent',
    'also',
    'attribute',
    'be_in_state',
    'causes',
    'classified_by',
    'classifies',
    'co_agent_instrument',
    'co_agent_patient',
    'co_agent_result',
    'co_instrument_agent',
    'co_instrument_patient',
    'co_instrument_result',
    'co_patient_agent',
    'co_patient_instrument',
    'co_result_agent',
    'co_result_instrument',
    'co_role',
    'direction',
    'domain_region',
    'domain_topic',
    'exemplifies',
    'entails',
    'eq_synonym',
    'has_domain_region',
    'has_domain_topic',
    'is_exemplified_by',
    'holo_location',
    'holo_member',
    'holo_part',
    'holo_portion',
    'holo_substance',
    'holonym',
    'hypernym',
    'hyponym',
    'in_manner',
    'instance_hypernym',
    'instance_hyponym',
    'instrument',
    'involved',
    'involved_agent',
    'involved_direction',
    'involved_instrument',
    'involved_location',
    'involved_patient',
    'involved_result',
    'involved_source_direction',
    'involved_target_direction',
    'is_caused_by',
    'is_entailed_by',
    'location',
    'manner_of',
    'mero_location',
    'mero_member',
    'mero_part',
    'mero_portion',
    'mero_substance',
    'meronym',
    'similar',
    'other',
    'patient',
    'restricted_by',
    'restricts',
    'result',
    'role',
    'source_direction',
    'state_of',
    'target_direction',
    'subevent',
    'is_subevent_of',
    'antonym',
    'feminine',
    'has_feminine',
    'masculine',
    'has_masculine',
    'young',
    'has_young',
    'diminutive',
    'has_diminutive',
    'augmentative',
    'has_augmentative',
    'anto_gradable',
    'anto_simple',
    'anto_converse',
    'ir_synonym',
])


REVERSE_RELATIONS = {
    'hypernym': 'hyponym',
    'hyponym': 'hypernym',
    'instance_hypernym': 'instance_hyponym',
    'instance_hyponym': 'instance_hypernym',
    'antonym': 'antonym',
    'eq_synonym': 'eq_synonym',
    'similar': 'similar',
    'meronym': 'holonym',
    'holonym': 'meronym',
    'mero_location': 'holo_location',
    'holo_location': 'mero_location',
    'mero_member': 'holo_member',
    'holo_member': 'mero_member',
    'mero_part': 'holo_part',
    'holo_part': 'mero_part',
    'mero_portion': 'holo_portion',
    'holo_portion': 'mero_portion',
    'mero_substance': 'holo_substance',
    'holo_substance': 'mero_substance',
    'also': 'also',
    'state_of': 'be_in_state',
    'be_in_state': 'state_of',
    'causes': 'is_caused_by',
    'is_caused_by': 'causes',
    'subevent': 'is_subevent_of',
    'is_subevent_of': 'subevent',
    'manner_of': 'in_manner',
    'in_manner': 'manner_of',
    'attribute': 'attribute',
    'restricts': 'restricted_by',
    'restricted_by': 'restricts',
    'classifies': 'classified_by',
    'classified_by': 'classifies',
    'entails': 'is_entailed_by',
    'is_entailed_by': 'entails',
    'domain_topic': 'has_domain_topic',
    'has_domain_topic': 'domain_topic',
    'domain_region': 'has_domain_region',
    'has_domain_region': 'domain_region',
    'exemplifies': 'is_exemplified_by',
    'is_exemplified_by': 'exemplifies',
    'role': 'involved',
    'involved': 'role',
    'agent': 'involved_agent',
    'involved_agent': 'agent',
    'patient': 'involved_patient',
    'involved_patient': 'patient',
    'result': 'involved_result',
    'involved_result': 'result',
    'instrument': 'involved_instrument',
    'involved_instrument': 'instrument',
    'location': 'involved_location',
    'involved_location': 'location',
    'direction': 'involved_direction',
    'involved_direction': 'direction',
    'target_direction': 'involved_target_direction',
    'involved_target_direction': 'target_direction',
    'source_direction': 'involved_source_direction',
    'involved_source_direction': 'source_direction',
    'co_role': 'co_role',
    'co_agent_patient': 'co_patient_agent',
    'co_patient_agent': 'co_agent_patient',
    'co_agent_instrument': 'co_instrument_agent',
    'co_instrument_agent': 'co_agent_instrument',
    'co_agent_result': 'co_result_agent',
    'co_result_agent': 'co_agent_result',
    'co_patient_instrument': 'co_instrument_patient',
    'co_instrument_patient': 'co_patient_instrument',
    'co_result_instrument': 'co_instrument_result',
    'co_instrument_result': 'co_result_instrument',
    'pertainym': 'pertainym',
    'derivation': 'derivation',
    'simple_aspect_ip': 'simple_aspect_pi',
    'simple_aspect_pi': 'simple_aspect_ip',
    'secondary_aspect_ip': 'secondary_aspect_pi',
    'secondary_aspect_pi': 'secondary_aspect_ip',
    'feminine': 'has_feminine',
    'has_feminine': 'feminine',
    'masculine': 'has_masculine',
    'has_masculine': 'masculine',
    'young': 'has_young',
    'has_young': 'young',
    'diminutive': 'has_diminutive',
    'has_diminutive': 'diminutive',
    'augmentative': 'has_augmentative',
    'has_augmentative': 'augmentative',
    'anto_gradable': 'anto_gradable',
    'anto_simple': 'anto_simple',
    'anto_converse': 'anto_converse',
    'ir_synonym': 'ir_synonym',
    # 'participle': '',
    # 'other': '',
}

# Adjective Positions

ADJPOSITIONS = frozenset((
    'a',   # attributive
    'ip',  # immediate postnominal
    'p',   # predicative
))


# Parts of Speech

NOUN = 'n'  #:
VERB = 'v'  #:
ADJ = ADJECTIVE = 'a'  #:
ADV = ADVERB = 'r'  #:
ADJ_SAT = ADJECTIVE_SATELLITE = 's'  #:
PHRASE = 't'  #:
CONJ = CONJUNCTION = 'c'  #:
ADP = ADPOSITION = 'p'  #:
OTHER = 'x'  #:
UNKNOWN = 'u'  #:

PARTS_OF_SPEECH = frozenset((
    NOUN,
    VERB,
    ADJECTIVE,
    ADVERB,
    ADJECTIVE_SATELLITE,
    PHRASE,
    CONJUNCTION,
    ADPOSITION,
    OTHER,
    UNKNOWN,
))


# Lexicographer Files
# from https://wordnet.princeton.edu/documentation/lexnames5wn

LEXICOGRAPHER_FILES = {
    'adj.all': 0,
    'adj.pert': 1,
    'adv.all': 2,
    'noun.Tops': 3,
    'noun.act': 4,
    'noun.animal': 5,
    'noun.artifact': 6,
    'noun.attribute': 7,
    'noun.body': 8,
    'noun.cognition': 9,
    'noun.communication': 10,
    'noun.event': 11,
    'noun.feeling': 12,
    'noun.food': 13,
    'noun.group': 14,
    'noun.location': 15,
    'noun.motive': 16,
    'noun.object': 17,
    'noun.person': 18,
    'noun.phenomenon': 19,
    'noun.plant': 20,
    'noun.possession': 21,
    'noun.process': 22,
    'noun.quantity': 23,
    'noun.relation': 24,
    'noun.shape': 25,
    'noun.state': 26,
    'noun.substance': 27,
    'noun.time': 28,
    'verb.body': 29,
    'verb.change': 30,
    'verb.cognition': 31,
    'verb.communication': 32,
    'verb.competition': 33,
    'verb.consumption': 34,
    'verb.contact': 35,
    'verb.creation': 36,
    'verb.emotion': 37,
    'verb.motion': 38,
    'verb.perception': 39,
    'verb.possession': 40,
    'verb.social': 41,
    'verb.stative': 42,
    'verb.weather': 43,
    'adj.ppl': 44,
}

# resource types

_WORDNET = 'wordnet'
_ILI = 'ili'




import pytest

import wn
from wn.taxonomy import (
    roots,
    leaves,
    taxonomy_depth,
    hypernym_paths,
    min_depth,
    max_depth,
    shortest_path,
    common_hypernyms,
    lowest_common_hypernyms,
)


@pytest.mark.usefixtures('mini_db')
def test_roots():
    en = wn.Wordnet('test-en')
    assert set(roots(en, pos='n')) == {en.synset('test-en-0001-n'),
                                       en.synset('test-en-0008-n')}
    assert set(roots(en, pos='v')) == {en.synset('test-en-0003-v'),
                                       en.synset('test-en-0007-v')}
    assert roots(en, pos='a') == []
    assert set(roots(en)) == set(roots(en, pos='n') + roots(en, pos='v'))

    # with no expand relations and no relation of its own, every
    # synset looks like a root
    es = wn.Wordnet('test-es')
    assert set(roots(es, pos='n')) == {es.synset('test-es-0001-n'),
                                       es.synset('test-es-0002-n'),
                                       es.synset('test-es-0005-n')}

    es = wn.Wordnet('test-es', expand='test-en')
    assert roots(es, pos='n') == [es.synset('test-es-0001-n')]


@pytest.mark.usefixtures('mini_db')
def test_leaves():
    en = wn.Wordnet('test-en')
    assert set(leaves(en, pos='n')) == {en.synset('test-en-0005-n'),
                                        en.synset('test-en-0006-n'),
                                        en.synset('test-en-0008-n')}
    assert set(leaves(en, pos='v')) == {en.synset('test-en-0003-v'),
                                        en.synset('test-en-0007-v')}


@pytest.mark.usefixtures('mini_db')
def test_taxonomy_depth():
    en = wn.Wordnet('test-en')
    assert taxonomy_depth(en, pos='n') == 3
    assert taxonomy_depth(en, pos='v') == 0


@pytest.mark.usefixtures('mini_db')
def test_hypernym_paths():
    information = wn.synsets('information')[0]
    example = wn.synsets('example')[0]
    sample = wn.synsets('sample')[0]
    random_sample = wn.synsets('random sample')[0]
    assert hypernym_paths(information) == []
    assert hypernym_paths(example) == [[information]]
    assert hypernym_paths(sample) == [[example, information]]
    assert hypernym_paths(random_sample) == [[sample, example, information]]


@pytest.mark.usefixtures('mini_db')
def test_interlingual_hypernym_paths():
    información = wn.synsets('información')[0]
    ejemplo = wn.synsets('ejemplo')[0]
    inferred = wn.Synset.empty('*INFERRED*')
    muestra_aleatoria = wn.synsets('muestra aleatoria')[0]
    assert hypernym_paths(información) == []
    assert hypernym_paths(ejemplo) == [[información]]
    assert hypernym_paths(muestra_aleatoria) == [[inferred, ejemplo, información]]


@pytest.mark.usefixtures('mini_db')
def test_shortest_path():
    information = wn.synsets('information')[0]
    example = wn.synsets('example')[0]
    sample = wn.synsets('sample')[0]
    random_sample = wn.synsets('random sample')[0]
    datum = wn.synsets('datum')[0]
    exemplify = wn.synsets('exemplify')[0]
    inferred_root = wn.Synset.empty('*INFERRED*')
    assert shortest_path(information, information) == []
    assert shortest_path(information, datum) == [datum]
    assert shortest_path(information, sample) == [example, sample]
    assert shortest_path(sample, information) == [example, information]
    assert shortest_path(random_sample, datum) == [sample, example, information, datum]
    with pytest.raises(wn.Error):
        shortest_path(example, exemplify)
    assert shortest_path(example, exemplify, simulate_root=True) == [
        information, inferred_root, exemplify
    ]


@pytest.mark.usefixtures('mini_db')
def test_min_depth():
    assert min_depth(wn.synsets('information')[0]) == 0
    assert min_depth(wn.synsets('example')[0]) == 1
    assert min_depth(wn.synsets('sample')[0]) == 2
    assert min_depth(wn.synsets('random sample')[0]) == 3


@pytest.mark.usefixtures('mini_db')
def test_max_depth():
    assert max_depth(wn.synsets('information')[0]) == 0
    assert max_depth(wn.synsets('example')[0]) == 1
    assert max_depth(wn.synsets('sample')[0]) == 2
    assert max_depth(wn.synsets('random sample')[0]) == 3




import pytest

import wn
from wn import morphy


def test_morphy_uninitialized():
    # An unintialized Morphy isn't very bright, but it starts up
    # fast. It relies on the database to filter bad items.
    m = morphy.Morphy()
    assert m('example', 'n') == {'n': {'example'}}
    assert m('examples', 'n') == {'n': {'examples', 'example'}}
    assert m('examples', 'v') == {'v': {'examples', 'example', 'exampl'}}
    assert m('exemplifying', 'n') == {'n': {'exemplifying'}}
    assert m('exemplifying', 'v') == {'v': {'exemplifying', 'exemplify', 'exemplifye'}}
    assert m('data', 'n') == {'n': {'data'}}
    assert m('datums', 'n') == {'n': {'datums', 'datum'}}  # expected false positive
    assert m('examples', None) == {None: {'examples'},
                                   'n': {'example'},
                                   'v': {'example', 'exampl'}}
    assert m('exemplifying', None) == {None: {'exemplifying'},
                                       'v': {'exemplify', 'exemplifye'}}
    assert m('data', None) == {None: {'data'}}


@pytest.mark.usefixtures('mini_db')
def test_morphy_initialized():
    w = wn.Wordnet('test-en:1')
    m = morphy.Morphy(wordnet=w)
    assert m('example', 'n') == {'n': {'example'}}
    assert m('examples', 'n') == {'n': {'example'}}
    assert m('examples', 'v') == {}
    assert m('exemplifying', 'n') == {}
    assert m('exemplifying', 'v') == {'v': {'exemplify'}}
    assert m('data', 'n') == {'n': {'datum'}}
    assert m('datums', 'n') == {'n': {'datum'}}  # expected false positive
    assert m('examples', None) == {'n': {'example'}}
    assert m('exemplifying', None) == {'v': {'exemplify'}}
    assert m('data', None) == {'n': {'datum'}}


@pytest.mark.usefixtures('mini_db')
def test_issue_154():
    # https://github.com/goodmami/wn/issues/154
    w = wn.Wordnet('test-en:1')
    assert w.words('exemplifies') == [w.word('test-en-exemplify-v')]
    assert w.words('samples') == []
    w = wn.Wordnet('test-en:1', lemmatizer=morphy.Morphy())
    assert w.words('exemplifies') == [w.word('test-en-exemplify-v')]
    assert w.words('samples') == [w.word('test-en-sample-n')]



"""
Adding and removing lexicons to/from the database.
"""

from typing import (
    Union, Optional, TypeVar, Type, Tuple, List, Dict, Set,
    Iterator, Iterable, Sequence, cast
)
from pathlib import Path
from itertools import islice
import sqlite3
import logging

import wn
from wn._types import AnyPath
from wn.constants import _WORDNET, _ILI
from wn._db import connect
from wn._queries import find_lexicons, get_lexicon_extensions, get_lexicon
from wn._util import normalize_form
from wn.util import ProgressHandler, ProgressBar
from wn.project import iterpackages
from wn import lmf
from wn import _ili


logger = logging.getLogger('wn')


BATCH_SIZE = 1000
DEFAULT_MEMBER_RANK = 127  # synset member rank when not specified by 'members'

ENTRY_QUERY = '''
    SELECT e.rowid
      FROM entries AS e
     WHERE e.id = ?
       AND e.lexicon_rowid = ?
'''
# forms don't have reliable ids, so also consider rank; this depends
# on each form having a unique rank, and this doesn't work for lexicon
# extensions
FORM_QUERY = '''
    SELECT f.rowid
      FROM forms AS f
      JOIN entries AS e ON f.entry_rowid = e.rowid
     WHERE e.id = ?
       AND e.lexicon_rowid = ?
       AND (f.id = ? OR f.rank = ?)
'''
SENSE_QUERY = '''
    SELECT s.rowid
      FROM senses AS s
     WHERE s.id = ?
       AND s.lexicon_rowid = ?
'''
SYNSET_QUERY = '''
    SELECT ss.rowid
      FROM synsets AS ss
     WHERE ss.id = ?
       AND ss.lexicon_rowid = ?
'''
RELTYPE_QUERY = '''
    SELECT rt.rowid
      FROM relation_types AS rt
     WHERE rt.type = ?
'''
ILISTAT_QUERY = '''
    SELECT ist.rowid
      FROM ili_statuses AS ist
     WHERE ist.status = ?
'''
LEXFILE_QUERY = '''
    SELECT lf.rowid
      FROM lexfiles AS lf
     WHERE lf.name = ?
'''

_AnyLexicon = Union[lmf.Lexicon, lmf.LexiconExtension]
_AnyEntry = Union[lmf.LexicalEntry, lmf.ExternalLexicalEntry]
_AnyLemma = Union[lmf.Lemma, lmf.ExternalLemma]
_AnyForm = Union[lmf.Form, lmf.ExternalForm]
_AnySense = Union[lmf.Sense, lmf.ExternalSense]
_AnySynset = Union[lmf.Synset, lmf.ExternalSynset]


def add(
    source: AnyPath,
    progress_handler: Optional[Type[ProgressHandler]] = ProgressBar,
) -> None:
    """Add the LMF file at *source* to the database.

    The file at *source* may be gzip-compressed or plain text XML.

    >>> wn.add('english-wordnet-2020.xml')
    Added ewn:2020 (English WordNet)

    The *progress_handler* parameter takes a subclass of
    :class:`wn.util.ProgressHandler`. An instance of the class will be
    created, used, and closed by this function.

    """
    if progress_handler is None:
        progress_handler = ProgressHandler
    progress = progress_handler(message='Database')

    logger.info('adding project to database')
    logger.info('  database: %s', wn.config.database_path)
    logger.info('  project file: %s', source)

    try:
        for package in iterpackages(source):
            if package.type == _WORDNET:
                _add_lmf(package.resource_file(), progress, progress_handler)
            elif package.type == _ILI:
                _add_ili(package.resource_file(), progress)
            else:
                raise wn.Error(f'unknown package type: {package.type}')
    finally:
        progress.close()


def _add_lmf(
    source: Path,
    progress: ProgressHandler,
    progress_handler: Type[ProgressHandler],
) -> None:
    with connect() as conn:
        cur = conn.cursor()
        # these two settings increase the risk of database corruption
        # if the system crashes during a write, but they should also
        # make inserts much faster
        cur.execute('PRAGMA synchronous = OFF')
        cur.execute('PRAGMA journal_mode = MEMORY')

        # abort if any lexicon in *source* is already added
        progress.flash(f'Checking {source!s}')
        all_infos = _precheck(source, cur)

        if not all_infos:
            progress.flash(f'{source}: No lexicons found')
            return
        for info in all_infos:
            skip = info.get('skip', '')
            if skip:
                id, ver, lbl = info['id'], info['version'], info['label']
                progress.flash(f'Skipping {id}:{ver} ({lbl}); {skip}\n')
        if all('skip' in info for info in all_infos):
            return

        # all clear, try to add them
        progress.flash(f'Reading {source!s}')
        resource = lmf.load(source, progress_handler)

        for lexicon, info in zip(resource['lexicons'], all_infos):
            if 'skip' in info:
                continue

            progress.flash('Updating lookup tables')
            _update_lookup_tables(lexicon, cur)

            progress.set(count=0, total=_sum_counts(lexicon))
            synsets: Sequence[_AnySynset] = _synsets(lexicon)
            entries: Sequence[_AnyEntry] = _entries(lexicon)
            synbhrs: Sequence[lmf.SyntacticBehaviour] = _collect_frames(lexicon)

            lexid, extid = _insert_lexicon(lexicon, info, cur, progress)

            lexidmap = _build_lexid_map(lexicon, lexid, extid)

            _insert_synsets(synsets, lexid, cur, progress)
            _insert_entries(entries, lexid, cur, progress)
            _insert_forms(entries, lexid, lexidmap, cur, progress)
            _insert_pronunciations(entries, lexid, lexidmap, cur, progress)
            _insert_tags(entries, lexid, lexidmap, cur, progress)
            _insert_senses(entries, synsets, lexid, lexidmap, cur, progress)
            _insert_adjpositions(entries, lexid, lexidmap, cur, progress)
            _insert_counts(entries, lexid, lexidmap, cur, progress)
            _insert_syntactic_behaviours(synbhrs, lexid, lexidmap, cur, progress)

            _insert_synset_relations(synsets, lexid, lexidmap, cur, progress)
            _insert_sense_relations(lexicon, lexid, lexidmap, cur, progress)

            _insert_synset_definitions(synsets, lexid, lexidmap, cur, progress)
            _insert_examples([sense for e in entries for sense in _senses(e)],
                             lexid, lexidmap, 'sense_examples', cur, progress)
            _insert_examples(synsets, lexid, lexidmap, 'synset_examples', cur, progress)

            progress.set(status='')  # clear type string
            progress.flash(
                f"Added {lexicon['id']}:{lexicon['version']} ({lexicon['label']})\n"
            )


def _precheck(source: Path, cur: sqlite3.Cursor) -> List[Dict]:
    lexqry = 'SELECT * FROM lexicons WHERE id = :id AND version = :version'
    infos = lmf.scan_lexicons(source)
    for info in infos:
        base = info.get('extends')
        if cur.execute(lexqry, info).fetchone():
            info['skip'] = 'already added'
        if base and cur.execute(lexqry, base).fetchone() is None:
            id_, ver = base['id'], base['version']
            info['skip'] = f'base lexicon ({id_}:{ver}) not available'
    return infos


def _sum_counts(lex: _AnyLexicon) -> int:
    ents = _entries(lex)
    locs = _local_entries(ents)
    lems = [e['lemma'] for e in locs if e.get('lemma')]
    frms = [f for e in ents for f in _forms(e)]
    sens = [s for e in ents for s in _senses(e)]
    syns = _synsets(lex)
    return sum([
        # lexical entries
        len(ents),
        len(lems),
        sum(len(lem.get('pronunciations', [])) for lem in lems),
        sum(len(lem.get('tags', [])) for lem in lems),
        len(frms),
        sum(len(frm.get('pronunciations', [])) for frm in frms),
        sum(len(frm.get('tags', [])) for frm in frms),
        # senses
        len(sens),
        sum(len(sen.get('relations', [])) for sen in sens),
        sum(len(sen.get('examples', [])) for sen in sens),
        sum(len(sen.get('counts', [])) for sen in sens),
        # synsets
        len(syns),
        sum(len(syn.get('definitions', [])) for syn in syns),
        sum(len(syn.get('relations', [])) for syn in syns),
        sum(len(syn.get('examples', [])) for syn in syns),
        # syntactic behaviours
        sum(len(ent.get('frames', [])) for ent in locs),
        len(lex.get('frames', [])),
    ])


def _update_lookup_tables(
    lexicon: _AnyLexicon,
    cur: sqlite3.Cursor
) -> None:
    reltypes = set(rel['relType']
                   for ss in _synsets(lexicon)
                   for rel in ss.get('relations', []))
    reltypes.update(rel['relType']
                    for e in _entries(lexicon)
                    for s in _senses(e)
                    for rel in s.get('relations', []))
    cur.executemany('INSERT OR IGNORE INTO relation_types VALUES (null,?)',
                    [(rt,) for rt in sorted(reltypes)])
    lexfiles: Set[str] = {ss.get('lexfile', '')
                          for ss in _local_synsets(_synsets(lexicon))
                          if ss.get('lexfile')}
    cur.executemany('INSERT OR IGNORE INTO lexfiles VALUES (null,?)',
                    [(lf,) for lf in sorted(lexfiles)])


def _insert_lexicon(
    lexicon: _AnyLexicon,
    info: dict,
    cur: sqlite3.Cursor,
    progress: ProgressHandler
) -> Tuple[int, int]:
    progress.set(status='Lexicon Info')
    cur.execute(
        'INSERT INTO lexicons VALUES (null,?,?,?,?,?,?,?,?,?,?,?)',
        (lexicon['id'],
         lexicon['label'],
         lexicon['language'],
         lexicon['email'],
         lexicon['license'],
         lexicon['version'],
         lexicon.get('url'),
         lexicon.get('citation'),
         lexicon.get('logo'),
         lexicon.get('meta'),
         False))
    lexid = cur.lastrowid

    query = '''
        UPDATE lexicon_dependencies
           SET provider_rowid = ?
         WHERE provider_id = ? AND provider_version = ?
    '''
    cur.execute(query, (lexid, lexicon['id'], lexicon['version']))

    query = '''
        INSERT INTO {table}
        VALUES (:lid,
                :id,
                :version,
                :url,
                (SELECT rowid FROM lexicons WHERE id=:id AND version=:version))
    '''
    params = []
    for dep in lexicon.get('requires', []):
        param_dict = dict(dep)
        param_dict.setdefault('url', None)
        param_dict['lid'] = lexid
        params.append(param_dict)
    if params:
        cur.executemany(query.format(table='lexicon_dependencies'), params)

    if lexicon.get('extends'):
        lexicon = cast(lmf.LexiconExtension, lexicon)
        param_dict = dict(lexicon['extends'])
        param_dict.setdefault('url', None)
        param_dict['lid'] = lexid
        cur.execute(query.format(table='lexicon_extensions'), param_dict)
        extid = cur.execute(
            'SELECT rowid FROM lexicons WHERE id=? AND version=?',
            (param_dict['id'], param_dict['version'])
        ).fetchone()[0]
    else:
        extid = lexid

    return lexid, extid


_LexIdMap = Dict[str, int]


def _build_lexid_map(lexicon: _AnyLexicon, lexid: int, extid: int) -> _LexIdMap:
    """Build a mapping of entity IDs to extended lexicon rowid."""
    lexidmap: _LexIdMap = {}
    if lexid != extid:
        lexidmap.update((e['id'], extid)
                        for e in _entries(lexicon)
                        if _is_external(e))
        lexidmap.update((s['id'], extid)
                        for e in _entries(lexicon)
                        for s in _senses(e)
                        if _is_external(s))
        lexidmap.update((ss['id'], extid)
                        for ss in _synsets(lexicon)
                        if _is_external(ss))
    return lexidmap


T = TypeVar('T')


def _batch(sequence: Iterable[T]) -> Iterator[List[T]]:
    it = iter(sequence)
    batch = list(islice(it, 0, BATCH_SIZE))
    while len(batch):
        yield batch
        batch = list(islice(it, 0, BATCH_SIZE))


def _insert_synsets(
    synsets: Sequence[_AnySynset],
    lexid: int,
    cur: sqlite3.Cursor,
    progress: ProgressHandler
) -> None:
    progress.set(status='Synsets')
    # synsets
    ss_query = f'''
        INSERT INTO synsets
        VALUES (null,?,?,(SELECT rowid FROM ilis WHERE id=?),?,?,({LEXFILE_QUERY}),?)
    '''
    # presupposed ILIs
    pre_ili_query = f'''
        INSERT OR IGNORE INTO ilis
        VALUES (null,?,({ILISTAT_QUERY}),?,?)
    '''
    # proposed ILIs
    pro_ili_query = '''
        INSERT INTO proposed_ilis
        VALUES (null,
               (SELECT ss.rowid
                  FROM synsets AS ss
                 WHERE ss.id=? AND lexicon_rowid=?),
               ?,
               ?)
    '''

    for batch in _batch(_local_synsets(synsets)):

        # first add presupposed ILIs
        pre_ili_data = []
        for ss in batch:
            ili = ss['ili']
            if ili and ili != 'in':
                defn = ss.get('ili_definition')  # normally null
                text = defn['text'] if defn else None
                meta = defn['meta'] if defn else None
                pre_ili_data.append((ili, 'presupposed', text, meta))
        cur.executemany(pre_ili_query, pre_ili_data)

        # then add synsets
        ss_data = (
            (ss['id'],
             lexid,
             ss['ili'] if ss['ili'] and ss['ili'] != 'in' else None,
             ss['partOfSpeech'],
             ss.get('lexicalized', True),
             ss.get('lexfile'),
             ss['meta'])
            for ss in batch
        )
        cur.executemany(ss_query, ss_data)

        # finally add proposed ILIs
        pro_ili_data = []
        for ss in batch:
            ili = ss['ili']
            if ili == 'in':
                defn = ss.get('ili_definition')
                text = defn['text'] if defn else None
                meta = defn['meta'] if defn else None
                pro_ili_data.append((ss['id'], lexid, text, meta))
        cur.executemany(pro_ili_query, pro_ili_data)

        progress.update(len(batch))


def _insert_synset_definitions(
    synsets: Sequence[_AnySynset],
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler
) -> None:
    progress.set(status='Definitions')
    query = f'''
        INSERT INTO definitions
        VALUES (null,?,({SYNSET_QUERY}),?,?,({SENSE_QUERY}),?)
    '''
    for batch in _batch(synsets):
        data = [
            (lexid,
             synset['id'],
             lexidmap.get(synset['id'], lexid),
             definition['text'],
             definition.get('language'),
             definition.get('sourceSense'),
             lexidmap.get(definition.get('sourceSense', ''), lexid),
             definition['meta'])
            for synset in batch
            for definition in synset.get('definitions', [])
        ]
        cur.executemany(query, data)
        progress.update(len(data))


def _insert_synset_relations(
    synsets: Sequence[_AnySynset],
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler
) -> None:
    progress.set(status='Synset Relations')
    query = f'''
        INSERT INTO synset_relations
        VALUES (null,?,({SYNSET_QUERY}),({SYNSET_QUERY}),({RELTYPE_QUERY}),?)
    '''
    for batch in _batch(synsets):
        data = [
            (lexid,
             synset['id'], lexidmap.get(synset['id'], lexid),
             relation['target'], lexidmap.get(relation['target'], lexid),
             relation['relType'],
             relation['meta'])
            for synset in batch
            for relation in synset.get('relations', [])
        ]
        cur.executemany(query, data)
        progress.update(len(data))


def _insert_entries(
    entries: Sequence[_AnyEntry],
    lexid: int,
    cur: sqlite3.Cursor,
    progress: ProgressHandler
) -> None:
    progress.set(status='Words')
    query = 'INSERT INTO entries VALUES (null,?,?,?,?)'
    for batch in _batch(_local_entries(entries)):
        data = (
            (entry['id'],
             lexid,
             entry['lemma']['partOfSpeech'],
             entry['meta'])
            for entry in batch
        )
        cur.executemany(query, data)
        progress.update(len(batch))


def _insert_forms(
    entries: Sequence[_AnyEntry],
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler
) -> None:
    progress.set(status='Word Forms')
    query = f'INSERT INTO forms VALUES (null,?,?,({ENTRY_QUERY}),?,?,?,?)'
    for batch in _batch(entries):
        forms: List[Tuple[Optional[str], int, str, int,
                          str, Optional[str], Optional[str], int]] = []
        for entry in batch:
            eid = entry['id']
            lid = lexidmap.get(eid, lexid)
            if not _is_external(entry):
                entry = cast(lmf.LexicalEntry, entry)
                written_form = entry['lemma']['writtenForm']
                norm = normalize_form(written_form)
                forms.append(
                    (None, lexid, eid, lid,
                     written_form, norm if norm != written_form else None,
                     entry['lemma'].get('script'), 0)
                )
            for i, form in enumerate(_forms(entry), 1):
                if _is_external(form):
                    continue
                form = cast(lmf.Form, form)
                written_form = form['writtenForm']
                norm = normalize_form(written_form)
                forms.append(
                    (form.get('id'), lexid, eid, lid,
                     written_form, norm if norm != written_form else None,
                     form.get('script'), i)
                )
        cur.executemany(query, forms)
        progress.update(len(forms))


def _insert_pronunciations(
    entries: Sequence[_AnyEntry],
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler
) -> None:
    progress.set(status='Pronunciations')
    query = f'INSERT INTO pronunciations VALUES (({FORM_QUERY}),?,?,?,?,?)'
    for batch in _batch(entries):
        prons: List[Tuple[str, int, Optional[str], int,
                          str, Optional[str], Optional[str],
                          bool, Optional[str]]] = []
        for entry in batch:
            eid = entry['id']
            lid = lexidmap.get(eid, lexid)
            if entry.get('lemma'):
                for p in entry['lemma'].get('pronunciations', []):
                    prons.append(
                        (eid, lid, None, 0,
                         p['text'], p.get('variety'), p.get('notation'),
                         p.get('phonemic', True), p.get('audio'))
                    )
            for i, form in enumerate(_forms(entry), 1):
                # rank is not valid in FORM_QUERY for external forms
                rank = -1 if _is_external(form) else i
                for p in form.get('pronunciations', []):
                    prons.append(
                        (eid, lid, form.get('id'), rank,
                         p['text'], p.get('variety'), p.get('notation'),
                         p.get('phonemic', True), p.get('audio'))
                    )
        cur.executemany(query, prons)
        progress.update(len(prons))


def _insert_tags(
    entries: Sequence[_AnyEntry],
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler
) -> None:
    progress.set(status='Word Form Tags')
    query = f'INSERT INTO tags VALUES (({FORM_QUERY}),?,?)'
    for batch in _batch(entries):
        tags: List[Tuple[str, int, Optional[str], int, str, str]] = []
        for entry in batch:
            eid = entry['id']
            lid = lexidmap.get(eid, lexid)
            if entry.get('lemma'):
                for tag in entry['lemma'].get('tags', []):
                    tags.append((eid, lid, None, 0, tag['text'], tag['category']))
            for i, form in enumerate(_forms(entry), 1):
                # rank is not valid in FORM_QUERY for external forms
                rank = -1 if _is_external(form) else i
                for tag in form.get('tags', []):
                    tags.append(
                        (eid, lid, form.get('id'), rank, tag['text'], tag['category'])
                    )
        cur.executemany(query, tags)
        progress.update(len(tags))


def _insert_senses(
    entries: Sequence[_AnyEntry],
    synsets: Sequence[_AnySynset],
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler
) -> None:
    progress.set(status='Senses')
    ssrank = {s: i
              for ss in _local_synsets(synsets)
              for i, s in enumerate(ss.get('members', []))}
    query = f'''
        INSERT INTO senses
        VALUES (null,
                ?,
                ?,
                ({ENTRY_QUERY}),
                ?,
                ({SYNSET_QUERY}),
                ?,
                ?,
                ?)
    '''
    for batch in _batch(entries):
        data = [
            (sense['id'],
             lexid,
             entry['id'], lexidmap.get(entry['id'], lexid),
             i,
             sense['synset'], lexidmap.get(sense['synset'], lexid),
             ssrank.get(sense['id'], DEFAULT_MEMBER_RANK),
             sense.get('lexicalized', True),
             sense['meta'])
            for entry in batch
            for i, sense in enumerate(_local_senses(_senses(entry)))
        ]
        cur.executemany(query, data)
        progress.update(len(data))


def _insert_adjpositions(
    entries: Sequence[_AnyEntry],
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler
):
    progress.set(status='Sense Adjpositions')
    data = [(s['id'], lexidmap.get(s['id'], lexid), s['adjposition'])
            for e in entries
            for s in _local_senses(_senses(e))
            if s.get('adjposition')]
    query = f'INSERT INTO adjpositions VALUES (({SENSE_QUERY}),?)'
    cur.executemany(query, data)


def _insert_counts(
    entries: Sequence[_AnyEntry],
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler
) -> None:
    progress.set(status='Counts')
    data = [(lexid,
             sense['id'], lexidmap.get(sense['id'], lexid),
             count['value'],
             count['meta'])
            for entry in entries
            for sense in _senses(entry)
            for count in sense.get('counts', [])]
    query = f'INSERT INTO counts VALUES (null,?,({SENSE_QUERY}),?,?)'
    cur.executemany(query, data)
    progress.update(len(data))


def _collect_frames(lexicon: _AnyLexicon) -> List[lmf.SyntacticBehaviour]:
    # WN-LMF 1.0 syntactic behaviours are on lexical entries, and in
    # WN-LMF 1.1 they are at the lexticon level with IDs. This
    # function normalizes the two variants.

    # IDs are not required and frame strings must be unique in a
    # lexicon, so lookup syntactic behaviours by the frame string
    synbhrs: Dict[str, lmf.SyntacticBehaviour] = {
        frame['subcategorizationFrame']: {
            'id': frame['id'],
            'subcategorizationFrame': frame['subcategorizationFrame'],
        }
        for frame in lexicon.get('frames', [])
    }
    # all relevant senses are collected into the 'senses' key
    id_senses_map = {sb['id']: sb.get('senses', [])
                     for sb in synbhrs.values() if sb.get('id')}
    for entry in _entries(lexicon):
        # for WN-LMF 1.1
        for sense in _local_senses(_senses(entry)):
            for sbid in sense.get('subcat', []):
                id_senses_map[sbid].append(sense['id'])
        # for WN-LMF 1.0
        if _is_external(entry) or not entry.get('frames'):
            continue
        entry = cast(lmf.LexicalEntry, entry)
        all_senses = [s['id'] for s in _senses(entry)]
        for frame in entry.get('frames', []):
            subcat_frame = frame['subcategorizationFrame']
            if subcat_frame not in synbhrs:
                synbhrs[subcat_frame] = {'subcategorizationFrame': subcat_frame}
            senses = frame.get('senses', []) or all_senses
            synbhrs[subcat_frame].setdefault('senses', []).extend(senses)
    return list(synbhrs.values())


def _insert_syntactic_behaviours(
    synbhrs: Sequence[lmf.SyntacticBehaviour],
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler
) -> None:
    progress.set(status='Syntactic Behaviours')

    query = 'INSERT INTO syntactic_behaviours VALUES (null,?,?,?)'
    sbdata = [(sb.get('id') or None, lexid, sb['subcategorizationFrame'])
              for sb in synbhrs]
    cur.executemany(query, sbdata)

    # syntactic behaviours don't have a required ID; index on frame
    framemap: Dict[str, List[str]] = {
        sb['subcategorizationFrame']: sb.get('senses', []) for sb in synbhrs
    }
    query = f'''
        INSERT INTO syntactic_behaviour_senses
        VALUES ((SELECT rowid
                   FROM syntactic_behaviours
                  WHERE lexicon_rowid=? AND frame=?),
                ({SENSE_QUERY}))
    '''
    sbsdata = [(lexid, frame, sid, lexidmap.get(sid, lexid))
               for frame in framemap
               for sid in framemap[frame]]
    cur.executemany(query, sbsdata)

    progress.update(len(synbhrs))


def _insert_sense_relations(
    lexicon: _AnyLexicon,
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler
) -> None:
    progress.set(status='Sense Relations')
    # need to separate relations into those targeting senses vs synsets
    synset_ids = {ss['id'] for ss in _synsets(lexicon)}
    sense_ids = {s['id']
                 for e in _entries(lexicon)
                 for s in _senses(e)}
    s_s_rels = []
    s_ss_rels = []
    for entry in _entries(lexicon):
        for sense in _senses(entry):
            slid = lexidmap.get(sense['id'], lexid)
            for relation in sense.get('relations', []):
                target_id = relation['target']
                tlid = lexidmap.get(target_id, lexid)
                if target_id in sense_ids:
                    s_s_rels.append((sense['id'], slid, tlid, relation))
                elif target_id in synset_ids:
                    s_ss_rels.append((sense['id'], slid, tlid, relation))
                else:
                    raise wn.Error(
                        f'relation target is not a known sense or synset: {target_id}'
                    )
    hyperparams = [
        ('sense_relations', SENSE_QUERY, s_s_rels),
        ('sense_synset_relations', SYNSET_QUERY, s_ss_rels),
    ]
    for table, target_query, rels in hyperparams:
        query = f'''
            INSERT INTO {table}
            VALUES (null,?,({SENSE_QUERY}),({target_query}),({RELTYPE_QUERY}),?)
        '''
        for batch in _batch(rels):
            data = [
                (lexid,
                 sense_id, slid,
                 relation['target'], tlid,
                 relation['relType'],
                 relation['meta'])
                for sense_id, slid, tlid, relation in batch
            ]
            cur.executemany(query, data)
            progress.update(len(data))


def _insert_examples(
    objs: Sequence[Union[lmf.Sense, lmf.ExternalSense, lmf.Synset, lmf.ExternalSynset]],
    lexid: int,
    lexidmap: _LexIdMap,
    table: str,
    cur: sqlite3.Cursor,
    progress: ProgressHandler
) -> None:
    progress.set(status='Examples')
    if table == 'sense_examples':
        query = f'INSERT INTO {table} VALUES (null,?,({SENSE_QUERY}),?,?,?)'
    else:
        query = f'INSERT INTO {table} VALUES (null,?,({SYNSET_QUERY}),?,?,?)'
    for batch in _batch(objs):
        data = [
            (lexid,
             obj['id'], lexidmap.get(obj['id'], lexid),
             example['text'],
             example.get('language'),
             example['meta'])
            for obj in batch
            for example in obj.get('examples', [])
        ]
        # be careful of SQL injection here
        cur.executemany(query, data)
        progress.update(len(data))


def _add_ili(
    source: Path,
    progress: ProgressHandler,
) -> None:
    query = f'''
        INSERT INTO ilis
        VALUES (null,?,({ILISTAT_QUERY}),?,null)
            ON CONFLICT(id) DO
               UPDATE SET status_rowid=excluded.status_rowid,
                          definition=excluded.definition
    '''
    with connect() as conn:
        cur = conn.cursor()

        progress.flash(f'Reading ILI file: {source!s}')
        ili = list(_ili.load(source))

        progress.flash('Updating ILI Status Names')
        statuses = set(info.get('status', 'active') for info in ili)
        cur.executemany('INSERT OR IGNORE INTO ili_statuses VALUES (null,?)',
                        [(stat,) for stat in sorted(statuses)])

        progress.set(count=0, total=len(ili), status='ILI')
        for batch in _batch(ili):
            data = [
                (info['ili'],
                 info.get('status', 'active'),
                 info.get('definition'))
                for info in batch
            ]
            cur.executemany(query, data)
            progress.update(len(data))


def remove(
    lexicon: str,
    progress_handler: Optional[Type[ProgressHandler]] = ProgressBar
) -> None:
    """Remove lexicon(s) from the database.

    The *lexicon* argument is a :ref:`lexicon specifier
    <lexicon-specifiers>`. Note that this removes a lexicon and not a
    project, so the lexicons of projects containing multiple lexicons
    will need to be removed individually or, if applicable, a star
    specifier.

    The *progress_handler* parameter takes a subclass of
    :class:`wn.util.ProgressHandler`. An instance of the class will be
    created, used, and closed by this function.

    >>> wn.remove('ewn:2019')  # removes a single lexicon
    >>> wn.remove('*:1.3+omw')  # removes all lexicons with version 1.3+omw

    """
    if progress_handler is None:
        progress_handler = ProgressHandler
    progress = progress_handler(message='Removing', unit='\be5 operations')

    conn = connect()
    conn.set_progress_handler(progress.update, 100000)
    try:
        for rowid, id, _, _, _, _, version, *_ in find_lexicons(lexicon=lexicon):
            extensions = _find_all_extensions(rowid)

            with conn:

                for ext_id, ext_spec in reversed(extensions):
                    progress.set(status=f'{ext_spec} (extension)')
                    conn.execute('DELETE from lexicons WHERE rowid = ?', (ext_id,))
                    progress.flash(f'Removed {ext_spec}\n')

                spec = f'{id}:{version}'
                extra = f' (and {len(extensions)} extension(s))' if extensions else ''
                progress.set(status=f'{spec}', count=0)
                conn.execute('DELETE from lexicons WHERE rowid = ?', (rowid,))
                progress.flash(f'Removed {spec}{extra}\n')

    finally:
        progress.close()
        conn.set_progress_handler(None, 0)


def _find_all_extensions(rowid: int) -> List[Tuple[int, str]]:
    exts: List[Tuple[int, str]] = []
    for ext_id in get_lexicon_extensions(rowid):
        lexinfo = get_lexicon(ext_id)
        exts.append((ext_id, f'{lexinfo[1]}:{lexinfo[6]}'))
    return exts


def _entries(lex: _AnyLexicon) -> Sequence[_AnyEntry]: return lex.get('entries', [])
def _forms(e: _AnyEntry) -> Sequence[_AnyForm]: return e.get('forms', [])
def _senses(e: _AnyEntry) -> Sequence[_AnySense]: return e.get('senses', [])
def _synsets(lex: _AnyLexicon) -> Sequence[_AnySynset]: return lex.get('synsets', [])


def _is_external(
    x: Union[_AnyForm, _AnyLemma, _AnyEntry, _AnySense, _AnySynset]
) -> bool:
    return x.get('external', False) is True


def _local_synsets(synsets: Sequence[_AnySynset]) -> Iterator[lmf.Synset]:
    for ss in synsets:
        if _is_external(ss):
            continue
        yield cast(lmf.Synset, ss)


def _local_entries(entries: Sequence[_AnyEntry]) -> Iterator[lmf.LexicalEntry]:
    for e in entries:
        if _is_external(e):
            continue
        yield cast(lmf.LexicalEntry, e)


def _local_senses(senses: Sequence[_AnySense]) -> Iterator[lmf.Sense]:
    for s in senses:
        if _is_external(s):
            continue
        yield cast(lmf.Sense, s)




import pytest

import wn


@pytest.mark.usefixtures('mini_db_1_1')
def test_wordnet_lexicons():
    en = wn.Wordnet('test-en')
    assert len(en.lexicons()) == 1
    assert len(en.expanded_lexicons()) == 0

    en1 = wn.Wordnet('test-en:1')
    assert en.lexicons() == en1.lexicons()
    assert en.expanded_lexicons() == en1.expanded_lexicons()

    en2 = wn.Wordnet(lang='en')
    assert len(en2.lexicons()) == 2
    assert len(en2.expanded_lexicons()) == 0

    es = wn.Wordnet('test-es')
    assert len(es.lexicons()) == 1
    assert len(es.expanded_lexicons()) == 0

    es2 = wn.Wordnet('test-es', expand='test-en')
    assert len(es2.lexicons()) == 1
    assert len(es2.expanded_lexicons()) == 1

    ja = wn.Wordnet('test-ja')
    assert len(ja.lexicons()) == 1
    assert len(ja.expanded_lexicons()) == 1

    ja2 = wn.Wordnet('test-ja', expand='')
    assert len(ja2.lexicons()) == 1
    assert len(ja2.expanded_lexicons()) == 0


@pytest.mark.usefixtures('mini_db')
def test_wordnet_normalize():
    es = wn.Wordnet('test-es')
    assert es.words('Informacion') == es.words('información')
    assert es.words('ínfórmácíón') == es.words('información')
    es = wn.Wordnet('test-es', normalizer=None)
    assert es.words('informacion') == []
    assert es.words('Información') == []

    # The following doesn't necessarily work because any non-None
    # normalizer causes the normalized form column to be tested with
    # the original form
    # es = wn.Wordnet('test-es', normalizer=str.lower)
    # assert es.words('informacion') == []
    # assert es.words('Información') == es.words('información')


@pytest.mark.usefixtures('mini_db')
def test_wordnet_lemmatize():
    # default lemmatizer compares alternative forms
    en = wn.Wordnet('test-en')
    assert en.words('examples') == []
    assert en.words('exemplifying') == en.words('exemplify')
    assert en.words('data') == en.words('datum')

    en = wn.Wordnet('test-en', search_all_forms=False)
    assert en.words('examples') == []
    assert en.words('exemplifying') == []
    assert en.words('data') == []

    def morphy_lite(form, pos):
        result = {pos: {form}}
        if pos in ('n', None) and form.endswith('s'):
            result.setdefault('n', set()).add(form[:-1])
        return result

    en = wn.Wordnet('test-en', lemmatizer=morphy_lite, search_all_forms=False)
    assert en.words('examples', pos='n') == en.words('example')
    assert en.words('examples') == en.words('example')
    assert en.words('exemplifying') == []
    assert en.words('data') == []

    en = wn.Wordnet('test-en', lemmatizer=morphy_lite, search_all_forms=True)
    assert en.words('data') == en.words('datum')
    assert en.words('exemplifying') == en.words('exemplify')




from math import log

import pytest

import wn
from wn.constants import (NOUN, VERB, ADJ, ADV)
from wn.util import synset_id_formatter
import wn.ic


synset_id = {
    'information': 'test-en-0001-n',
    'illustration_example': 'test-en-0002-n',
    'sample': 'test-en-0004-n',
    'random_sample': 'test-en-0005-n',
    'random_sample2': 'test-en-0008-n',  # no hypernyms
    'datum': 'test-en-0006-n',
    'illustrate_exemplify': 'test-en-0003-v',
    'resignate': 'test-en-0007-v',
}


words = [
    'For', 'example', ':', 'random sample', '.',
    'This', 'will', 'illustrate', 'and', 'exemplify', '.',
    'A', 'sample', 'of', 'data', '.',
]


@pytest.mark.usefixtures('mini_db')
def test_compute_nodistribute_nosmoothing():
    w = wn.Wordnet('test-en:1')
    assert wn.ic.compute(words, w, distribute_weight=False, smoothing=0) == {
        NOUN: {
            synset_id['information']: 4.0,
            synset_id['illustration_example']: 3.0,
            synset_id['sample']: 2.0,
            synset_id['random_sample']: 1.0,
            synset_id['random_sample2']: 1.0,
            synset_id['datum']: 1.0,
            None: 5.0,
        },
        VERB: {
            synset_id['illustrate_exemplify']: 2.0,
            synset_id['resignate']: 0.0,
            None: 2.0,
        },
        ADJ: {None: 0.0},
        ADV: {None: 0.0},
    }


@pytest.mark.usefixtures('mini_db')
def test_compute_nodistribute_smoothing():
    w = wn.Wordnet('test-en:1')
    assert wn.ic.compute(words, w, distribute_weight=False, smoothing=1.0) == {
        NOUN: {
            synset_id['information']: 5.0,
            synset_id['illustration_example']: 4.0,
            synset_id['sample']: 3.0,
            synset_id['random_sample']: 2.0,
            synset_id['random_sample2']: 2.0,
            synset_id['datum']: 2.0,
            None: 6.0,
        },
        VERB: {
            synset_id['illustrate_exemplify']: 3.0,
            synset_id['resignate']: 1.0,
            None: 3.0,
        },
        ADJ: {None: 1.0},
        ADV: {None: 1.0},
    }


@pytest.mark.usefixtures('mini_db')
def test_compute_distribute_smoothing():
    w = wn.Wordnet('test-en:1')
    assert wn.ic.compute(words, w, distribute_weight=True, smoothing=1.0) == {
        NOUN: {
            synset_id['information']: 4.5,
            synset_id['illustration_example']: 3.5,
            synset_id['sample']: 2.5,
            synset_id['random_sample']: 1.5,
            synset_id['random_sample2']: 1.5,
            synset_id['datum']: 2.0,
            None: 5.0,
        },
        VERB: {
            synset_id['illustrate_exemplify']: 3.0,
            synset_id['resignate']: 1.0,
            None: 3.0,
        },
        ADJ: {None: 1.0},
        ADV: {None: 1.0},
    }


@pytest.mark.usefixtures('mini_db')
def test_load(tmp_path):
    w = wn.Wordnet('test-en:1')
    icpath = tmp_path / 'foo.dat'
    icpath.write_text(
        'wnver:1234567890AbCdEf\n'
        '1n 4.0 ROOT\n'
        '2n 3.0\n'
        '4n 2.0\n'
        '5n 1.0\n'
        '8n 1.0 ROOT\n'
        '6n 1.0\n'
        '3v 2.0 ROOT\n'
        '7v 0.0 ROOT\n'
    )

    get_synset_id = synset_id_formatter('test-en-{offset:04}-{pos}')
    assert (wn.ic.load(icpath, w, get_synset_id=get_synset_id)
            == wn.ic.compute(words, w, distribute_weight=False, smoothing=0.0))


@pytest.mark.usefixtures('mini_db')
def test_information_content():
    w = wn.Wordnet('test-en:1')
    ic = wn.ic.compute(words, w)
    info = w.synsets('information')[0]
    samp = w.synsets('sample')[0]
    # info is a root but not the only one, so its IC is not 0.0
    assert wn.ic.information_content(info, ic) == -log(
        ic['n'][info.id]
        / ic['n'][None]
    )
    assert wn.ic.information_content(samp, ic) == -log(
        ic['n'][samp.id]
        / ic['n'][None]
    )




"""
Local configuration settings.
"""

from typing import Optional, Dict, Sequence, Any
from pathlib import Path

import tomli

from wn import ConfigurationError, ProjectError
from wn._types import AnyPath
from wn.constants import _WORDNET
from wn._util import resources, short_hash

# The directory where downloaded and added data will be stored.
DEFAULT_DATA_DIRECTORY = Path.home() / '.wn_data'
DATABASE_FILENAME = 'wn.db'


class WNConfig:

    def __init__(self):
        self._data_directory = DEFAULT_DATA_DIRECTORY
        self._projects = {}
        self._dbpath = self._data_directory / DATABASE_FILENAME
        self.allow_multithreading = False

    @property
    def data_directory(self) -> Path:
        """The file system directory where Wn's data is stored."""
        dir = self._data_directory
        dir.mkdir(exist_ok=True)
        return dir

    @data_directory.setter
    def data_directory(self, path):
        dir = Path(path).expanduser()
        if dir.exists() and not dir.is_dir():
            raise ConfigurationError(f'path exists and is not a directory: {dir}')
        self._data_directory = dir
        self._dbpath = dir / DATABASE_FILENAME

    @property
    def database_path(self):
        """The path to the database file."""
        return self._dbpath

    @property
    def downloads_directory(self):
        """The file system directory where downloads are cached."""
        dir = self.data_directory / 'downloads'
        dir.mkdir(exist_ok=True)
        return dir

    @property
    def index(self) -> Dict[str, Dict]:
        """The project index."""
        return self._projects

    def add_project(
        self,
        id: str,
        type: str = _WORDNET,
        label: str = None,
        language: str = None,
        license: str = None,
        error: str = None,
    ) -> None:
        """Add a new wordnet project to the index.

        Arguments:
            id: short identifier of the project
            type: project type (default 'wordnet')
            label: full name of the project
            language: `BCP 47`_ language code of the resource
            license: link or name of the project's default license
            error: if set, the error message to use when the project
              is accessed

        .. _BCP 47: https://en.wikipedia.org/wiki/IETF_language_tag
        """
        if id in self._projects:
            raise ValueError(f'project already added: {id}')
        self._projects[id] = {
            'type': type,
            'label': label,
            'language': language,
            'versions': {},
            'license': license,
        }
        if error:
            self._projects[id]['error'] = error

    def add_project_version(
        self,
        id: str,
        version: str,
        url: str = None,
        error: str = None,
        license: str = None,
    ) -> None:
        """Add a new resource version for a project.

        Exactly one of *url* or *error* must be specified.

        Arguments:
            id: short identifier of the project
            version: version string of the resource
            url: space-separated list of web addresses for the resource
            license: link or name of the resource's license; if not
              given, the project's default license will be used.
            error: if set, the error message to use when the project
              is accessed

        """
        version_data: Dict[str, Any]
        if url and not error:
            version_data = {'resource_urls': url.split()}
        elif error and not url:
            version_data = {'error': error}
        elif url and error:
            raise ConfigurationError(f'{id}:{version} specifies both url and redirect')
        else:
            version_data = {}
        if license:
            version_data['license'] = license
        project = self._projects[id]
        project['versions'][version] = version_data

    def get_project_info(self, arg: str) -> Dict:
        """Return information about an indexed project version.

        If the project has been downloaded and cached, the ``"cache"``
        key will point to the path of the cached file, otherwise its
        value is ``None``.

        Arguments:
            arg: a project specifier

        Example:

            >>> info = wn.config.get_project_info('oewn:2021')
            >>> info['label']
            'Open English WordNet'

        """
        id, _, version = arg.partition(':')
        if id not in self._projects:
            raise ProjectError(f'no such project id: {id}')
        project: Dict = self._projects[id]
        if 'error' in project:
            raise ProjectError(project['error'])

        versions: Dict = project['versions']
        if not version or version == '*':
            version = next(iter(versions), '')
        if not version:
            raise ProjectError(f'no versions available for {id}')
        elif version not in versions:
            raise ProjectError(f'no such version: {version!r} ({id})')
        info = versions[version]
        if 'error' in info:
            raise ProjectError(info['error'])

        urls = info.get('resource_urls', [])

        return dict(
            id=id,
            version=version,
            type=project['type'],
            label=project['label'],
            language=project['language'],
            license=info.get('license', project.get('license')),
            resource_urls=urls,
            cache=_get_cache_path_for_urls(self, urls),
        )

    def get_cache_path(self, url: str) -> Path:
        """Return the path for caching *url*.

        Note that in general this is just a path operation and does
        not signify that the file exists in the file system.

        """
        filename = short_hash(url)
        return self.downloads_directory / filename

    def update(self, data: dict) -> None:
        """Update the configuration with items in *data*.

        Items are only inserted or replaced, not deleted. If a project
        index is provided in the ``"index"`` key, then either the
        project must not already be indexed or any project fields
        (label, language, or license) that are specified must be equal
        to the indexed project.

        """
        if 'data_directory' in data:
            self.data_directory = data['data_directory']
        for id, project in data.get('index', {}).items():
            if id in self._projects:
                # validate that they are the same
                _project = self._projects[id]
                for attr in ('label', 'language', 'license'):
                    if attr in project and project[attr] != _project[attr]:
                        raise ConfigurationError(f'{attr} mismatch for {id}')
            else:
                self.add_project(
                    id,
                    type=project.get('type', _WORDNET),
                    label=project.get('label'),
                    language=project.get('language'),
                    license=project.get('license'),
                    error=project.get('error'),
                )
            for version, info in project.get('versions', {}).items():
                if 'url' in info and 'error' in project:
                    raise ConfigurationError(
                        f'{id}:{version} url specified with default error'
                    )
                self.add_project_version(
                    id,
                    version,
                    url=info.get('url'),
                    license=info.get('license'),
                    error=info.get('error'),
                )

    def load_index(self, path: AnyPath) -> None:
        """Load and update with the project index at *path*.

        The project index is a TOML_ file containing project and
        version information. For example:

        .. code-block:: toml

           [ewn]
             label = "Open English WordNet"
             language = "en"
             license = "https://creativecommons.org/licenses/by/4.0/"
             [ewn.versions.2019]
               url = "https://en-word.net/static/english-wordnet-2019.xml.gz"
             [ewn.versions.2020]
               url = "https://en-word.net/static/english-wordnet-2020.xml.gz"

        .. _TOML: https://toml.io

        """
        path = Path(path).expanduser()
        with path.open('rb') as indexfile:
            try:
                index = tomli.load(indexfile)
            except tomli.TOMLDecodeError as exc:
                raise ConfigurationError('malformed index file') from exc
        self.update({'index': index})


def _get_cache_path_for_urls(
    config: WNConfig,
    urls: Sequence[str],
) -> Optional[Path]:
    for url in urls:
        path = config.get_cache_path(url)
        if path.is_file():
            return path
    return None


config = WNConfig()
with resources.path('wn', 'index.toml') as index_path:
    config.load_index(index_path)




"""
Reader for the Lexical Markup Framework (LMF) format.
"""

from typing import (
    Type,
    List,
    Tuple,
    Dict,
    Optional,
    TextIO,
    BinaryIO,
    Any,
    Union,
    cast
)
import sys
if sys.version_info >= (3, 8):
    from typing import TypedDict, Literal
else:
    from typing_extensions import TypedDict, Literal
import re
from pathlib import Path
import xml.etree.ElementTree as ET  # for general XML parsing
import xml.parsers.expat  # for fast scanning of Lexicon versions
from xml.sax.saxutils import quoteattr

import wn
from wn._types import AnyPath, VersionInfo
from wn._util import is_xml, version_info
from wn.util import ProgressHandler, ProgressBar


class LMFError(wn.Error):
    """Raised on invalid LMF-XML documents."""


class LMFWarning(Warning):
    """Issued on non-conforming LFM values."""


SUPPORTED_VERSIONS = {'1.0', '1.1'}
_XMLDECL = b'<?xml version="1.0" encoding="UTF-8"?>'
_DOCTYPE = '<!DOCTYPE LexicalResource SYSTEM "{schema}">'
_SCHEMAS = {
    '1.0': 'http://globalwordnet.github.io/schemas/WN-LMF-1.0.dtd',
    '1.1': 'http://globalwordnet.github.io/schemas/WN-LMF-1.1.dtd',
}
_DOCTYPES = {
    _DOCTYPE.format(schema=schema): version for version, schema in _SCHEMAS.items()
}

_DC_URIS = {
    '1.0': 'http://purl.org/dc/elements/1.1/',
    '1.1': 'https://globalwordnet.github.io/schemas/dc/',
}
_DC_ATTRS = [
    'contributor',
    'coverage',
    'creator',
    'date',
    'description',
    'format',
    'identifier',
    'publisher',
    'relation',
    'rights',
    'source',
    'subject',
    'title',
    'type',
]
_NS_ATTRS = {
    version: dict(
        [(f'{uri} {attr}', attr) for attr in _DC_ATTRS]
        + [('status', 'status'),
           ('note', 'note'),
           ('confidenceScore', 'confidenceScore')]
    )
    for version, uri in _DC_URIS.items()
}

_LMF_1_0_ELEMS: Dict[str, str] = {
    'LexicalResource': 'lexical-resource',
    'Lexicon': 'lexicons',
    'LexicalEntry': 'entries',
    'Lemma': 'lemma',
    'Form': 'forms',
    'Tag': 'tags',
    'Sense': 'senses',
    'SenseRelation': 'relations',
    'Example': 'examples',
    'Count': 'counts',
    'SyntacticBehaviour': 'frames',
    'Synset': 'synsets',
    'Definition': 'definitions',
    'ILIDefinition': 'ili_definition',
    'SynsetRelation': 'relations',
}
_LMF_1_1_ELEMS = dict(_LMF_1_0_ELEMS)
_LMF_1_1_ELEMS.update({
    'Requires': 'requires',
    'Extends': 'extends',
    'Pronunciation': 'pronunciations',
    'LexiconExtension': 'lexicons',
    'ExternalLexicalEntry': 'entries',
    'ExternalLemma': 'lemma',
    'ExternalForm': 'forms',
    'ExternalSense': 'senses',
    'ExternalSynset': 'synsets',
})
_VALID_ELEMS = {
    '1.0': _LMF_1_0_ELEMS,
    '1.1': _LMF_1_1_ELEMS,
}
_LIST_ELEMS = {  # elements that collect into lists
    'Lexicon',
    'LexicalEntry',
    'Form',
    'Pronunciation',
    'Tag',
    'Sense',
    'SenseRelation',
    'Example',
    'Count',
    'Synset',
    'Definition',
    'SynsetRelation',
    'SyntacticBehaviour',
    'LexiconExtension',
    'Requires',
    'ExternalLexicalEntry',
    'ExternalForm',
    'ExternalSense',
    'ExternalSynset',
}
_CDATA_ELEMS = {  # elements with inner text
    'Pronunciation',
    'Tag',
    'Definition',
    'ILIDefinition',
    'Example',
    'Count',
}
_META_ELEMS = {  # elements with metadata
    'Lexicon',
    'LexicalEntry',
    'Sense',
    'SenseRelation',
    'Example',
    'Count',
    'Synset',
    'Definition',
    'ILIDefinition',
    'SynsetRelation',
    'LexiconExtension',
}


# WN-LMF Modeling ######################################################

# WN-LMF type-checking is handled via TypedDicts.  Inheritance and
# `total=False` are used to model optionality. For more information
# about this tactic, see https://www.python.org/dev/peps/pep-0589/.

class Metadata(TypedDict, total=False):
    contributor: str
    coverage: str
    creator: str
    date: str
    description: str
    format: str
    identifier: str
    publisher: str
    relation: str
    rights: str
    source: str
    subject: str
    title: str
    type: str
    status: str
    note: str
    confidenceScore: float


_HasId = TypedDict('_HasId', {'id': str})
_HasVersion = TypedDict('_HasVersion', {'version': str})
_HasILI = TypedDict('_HasILI', {'ili': str})
_HasSynset = TypedDict('_HasSynset', {'synset': str})
_MaybeId = TypedDict('_MaybeId', {'id': str}, total=False)
_HasText = TypedDict('_HasText', {'text': str})
_MaybeScript = TypedDict('_MaybeScript', {'script': str}, total=False)
_HasMeta = TypedDict('_HasMeta', {'meta': Optional[Metadata]})
_External = TypedDict('_External', {'external': Literal['true']})


class ILIDefinition(_HasText, _HasMeta):
    ...


class Definition(_HasText, _HasMeta, total=False):
    language: str
    sourceSense: str


class Relation(_HasMeta):
    target: str
    relType: str


class Example(_HasText, _HasMeta, total=False):
    language: str


class Synset(_HasId, _HasILI, _HasMeta, total=False):
    ili_definition: ILIDefinition
    partOfSpeech: str
    definitions: List[Definition]
    relations: List[Relation]
    examples: List[Example]
    lexicalized: bool
    members: List[str]
    lexfile: str


class ExternalSynset(_HasId, _External, total=False):
    definitions: List[Definition]
    relations: List[Relation]
    examples: List[Example]


class Count(_HasMeta):
    value: int


class Sense(_HasId, _HasSynset, _HasMeta, total=False):
    relations: List[Relation]
    examples: List[Example]
    counts: List[Count]
    lexicalized: bool
    adjposition: str
    subcat: List[str]


class ExternalSense(_HasId, _External, total=False):
    relations: List[Relation]
    examples: List[Example]
    counts: List[Count]


class Pronunciation(_HasText, total=False):
    variety: str
    notation: str
    phonemic: bool
    audio: str


class Tag(_HasText):
    category: str


class _FormChildren(TypedDict, total=False):
    pronunciations: List[Pronunciation]
    tags: List[Tag]


class Lemma(_MaybeScript, _FormChildren):
    writtenForm: str
    partOfSpeech: str


class ExternalLemma(_FormChildren, _External):
    ...


class Form(_MaybeId, _MaybeScript, _FormChildren):
    writtenForm: str


class ExternalForm(_HasId, _FormChildren, _External):
    ...


class _SyntacticBehaviourBase(_MaybeId):
    subcategorizationFrame: str


class SyntacticBehaviour(_SyntacticBehaviourBase, total=False):
    senses: List[str]


class _LexicalEntryBase(_HasId, _HasMeta, total=False):
    forms: List[Form]
    senses: List[Sense]
    frames: List[SyntacticBehaviour]


class LexicalEntry(_LexicalEntryBase):
    lemma: Lemma


class ExternalLexicalEntry(_HasId, _External, total=False):
    lemma: ExternalLemma
    forms: List[Union[Form, ExternalForm]]
    senses: List[Union[Sense, ExternalSense]]


class Dependency(_HasId, _HasVersion, total=False):
    url: str


class _LexiconBase(_HasMeta):
    id: str
    version: str
    label: str
    language: str
    email: str
    license: str


class Lexicon(_LexiconBase, total=False):
    url: str
    citation: str
    logo: str
    requires: List[Dependency]
    entries: List[LexicalEntry]
    synsets: List[Synset]
    frames: List[SyntacticBehaviour]


class LexiconExtension(_LexiconBase, total=False):
    url: str
    citation: str
    logo: str
    extends: Dependency
    requires: List[Dependency]
    entries: List[Union[LexicalEntry, ExternalLexicalEntry]]
    synsets: List[Union[Synset, ExternalSynset]]
    frames: List[SyntacticBehaviour]


class LexicalResource(TypedDict):
    lmf_version: str
    lexicons: List[Union[Lexicon, LexiconExtension]]


# Reading ##############################################################

def is_lmf(source: AnyPath) -> bool:
    """Return True if *source* is a WN-LMF file."""
    source = Path(source).expanduser()
    if not is_xml(source):
        return False
    with source.open(mode='rb') as fh:
        try:
            _read_header(fh)
        except LMFError:
            return False
    return True


def _read_header(fh: BinaryIO) -> str:
    xmldecl = fh.readline().rstrip().replace(b"'", b'"')
    doctype = fh.readline().rstrip().replace(b"'", b'"')

    if xmldecl != _XMLDECL:
        raise LMFError('invalid or missing XML declaration')

    # the XML declaration states that the file is UTF-8 (other
    # encodings are not allowed)
    doctype_decoded = doctype.decode('utf-8')
    if doctype_decoded not in _DOCTYPES:
        raise LMFError('invalid or missing DOCTYPE declaration')

    return _DOCTYPES[doctype_decoded]


def scan_lexicons(source: AnyPath) -> List[Dict]:
    """Scan *source* and return only the top-level lexicon info."""

    source = Path(source).expanduser()
    infos: List[Dict] = []

    lex_re = re.compile(b'<(Lexicon|LexiconExtension|Extends)\\b([^>]*)>', flags=re.M)
    attr_re = re.compile(b'''\\b(id|version|label)=["']([^"']+)["']''', flags=re.M)

    with open(source, 'rb') as fh:
        for m in lex_re.finditer(fh.read()):
            lextype, remainder = m.groups()
            info = {_m.group(1).decode('utf-8'): _m.group(2).decode('utf-8')
                    for _m in attr_re.finditer(remainder)}
            if 'id' not in info or 'version' not in info:
                raise LMFError(f'<{lextype.decode("utf-8")}> missing id or version')
            if lextype != b'Extends':
                infos.append(info)
            elif len(infos) > 0:
                infos[-1]['extends'] = info
            else:
                raise LMFError('invalid use of <Extends> in WN-LMF file')

    return infos


_Elem = Dict[str, Any]  # basic type for the loaded XML data


def load(
    source: AnyPath,
    progress_handler: Optional[Type[ProgressHandler]] = ProgressBar
) -> LexicalResource:
    """Load wordnets encoded in the WN-LMF format.

    Args:
        source: path to a WN-LMF file
    """
    source = Path(source).expanduser()
    if progress_handler is None:
        progress_handler = ProgressHandler

    version, num_elements = _quick_scan(source)
    progress = progress_handler(
        message='Read', total=num_elements, refresh_interval=10000
    )

    root: Dict[str, _Elem] = {}
    parser = _make_parser(root, version, progress)

    with open(source, 'rb') as fh:
        try:
            parser.ParseFile(fh)
        except xml.parsers.expat.ExpatError as exc:
            raise LMFError('invalid or ill-formed XML') from exc

    progress.close()

    resource: LexicalResource = {
        'lmf_version': version,
        'lexicons': [_validate(lex)
                     for lex in root['lexical-resource'].get('lexicons', [])],
    }

    return resource


def _quick_scan(source: Path) -> Tuple[str, int]:
    with source.open('rb') as fh:
        version = _read_header(fh)
        # _read_header() only reads the first 2 lines
        remainder = fh.read()
        num_elements = remainder.count(b'</') + remainder.count(b'/>')
    return version, num_elements


def _make_parser(root, version, progress):  # noqa: C901
    stack = [root]
    ELEMS = _VALID_ELEMS[version]
    NS_ATTRS = _NS_ATTRS[version]
    CDATA_ELEMS = _CDATA_ELEMS & set(ELEMS)
    LIST_ELEMS = _LIST_ELEMS & set(ELEMS)

    p = xml.parsers.expat.ParserCreate(namespace_separator=' ')
    has_text = False

    def start(name, attrs):
        nonlocal has_text
        if name in CDATA_ELEMS:
            has_text = True

        if name in _META_ELEMS:
            meta = {}
            for attr in list(attrs):
                if attr in NS_ATTRS:
                    meta[NS_ATTRS[attr]] = attrs.pop(attr)
            attrs['meta'] = meta or None

        if name.startswith('External'):
            attrs['external'] = True

        parent = stack[-1]
        key = ELEMS.get(name)
        if name in LIST_ELEMS:
            parent.setdefault(key, []).append(attrs)
        elif key is None or key in parent:
            raise _unexpected(name, p)
        else:
            parent[key] = attrs

        stack.append(attrs)

    def char_data(data):
        if has_text:
            parent = stack[-1]
            # sometimes the buffering occurs in the middle of text; if
            # so, append the current data
            if 'text' in parent:
                parent['text'] += data
            else:
                parent['text'] = data

    def end(name):
        nonlocal has_text
        has_text = False
        stack.pop()
        progress.update(force=(name == 'LexicalResource'))

    p.StartElementHandler = start
    p.EndElementHandler = end
    p.CharacterDataHandler = char_data

    return p


def _unexpected(name: str, p: xml.parsers.expat.XMLParserType) -> LMFError:
    return LMFError(f'unexpected element at line {p.CurrentLineNumber}: {name}')


# Validation ###########################################################

def _validate(elem: _Elem) -> Union[Lexicon, LexiconExtension]:
    ext = elem.get('extends')
    if ext:
        assert 'id' in ext
        assert 'version' in ext
        _validate_lexicon(elem, True)
        return cast(LexiconExtension, elem)
    else:
        _validate_lexicon(elem, False)
        return cast(Lexicon, elem)


def _validate_lexicon(elem: _Elem, extension: bool) -> None:
    for attr in 'id', 'version', 'label', 'language', 'email', 'license':
        assert attr in elem, f'<Lexicon> missing required attribute: {attr}'
    for dep in elem.get('requires', []):
        assert 'id' in dep
        assert 'version' in dep
    _validate_entries(elem.get('entries', []), extension)
    _validate_synsets(elem.get('synsets', []), extension)
    _validate_frames(elem.get('frames', []))


def _validate_entries(elems: List[_Elem], extension: bool) -> None:
    for elem in elems:
        assert 'id' in elem
        if not extension:
            assert not elem.get('external')
        lemma = elem.get('lemma')
        if not elem.get('external'):
            assert lemma is not None
            elem.setdefault('meta')
        # lemma and forms are the same except for partOfSpeech and id
        if lemma is not None and not lemma.get('external'):
            assert 'partOfSpeech' in lemma
        for form in elem.get('forms', []):
            assert not form.get('external') or form.get('id')
        _validate_forms(([lemma] if lemma else []) + elem.get('forms', []), extension)
        _validate_senses(elem.get('senses', []), extension)
        _validate_frames(elem.get('frames', []))


def _validate_forms(elems: List[_Elem], extension: bool) -> None:
    for elem in elems:
        if not extension:
            assert not elem.get('external')
        if not elem.get('external'):
            assert 'writtenForm' in elem
        for pron in elem.get('pronunciations', []):
            pron.setdefault('text', '')
            if pron.get('phonemic'):
                pron['phonemic'] = False if pron['phonemic'] == 'false' else True
        for tag in elem.get('tags', []):
            tag.setdefault('text', '')
            assert 'category' in tag


def _validate_senses(elems: List[_Elem], extension: bool) -> None:
    for elem in elems:
        assert 'id' in elem
        if not extension:
            assert not elem.get('external')
        if not elem.get('external'):
            assert 'synset' in elem
            elem.setdefault('meta')
        for rel in elem.get('relations', []):
            assert 'target' in rel
            assert 'relType' in rel
            rel.setdefault('meta')
        for ex in elem.get('examples', []):
            ex.setdefault('text', '')
            ex.setdefault('meta')
        for cnt in elem.get('counts', []):
            assert 'text' in cnt
            cnt['value'] = int(cnt.pop('text'))
            cnt.setdefault('meta')
        if elem.get('lexicalized'):
            elem['lexicalized'] = False if elem['lexicalized'] == 'false' else True
        if elem.get('subcat'):
            elem['subcat'] = elem['subcat'].split()


def _validate_frames(elems: List[_Elem]) -> None:
    for elem in elems:
        assert 'subcategorizationFrame' in elem
        if elem.get('senses'):
            elem['senses'] = elem['senses'].split()


def _validate_synsets(elems: List[_Elem], extension: bool) -> None:
    for elem in elems:
        assert 'id' in elem
        if not extension:
            assert not elem.get('external')
        if not elem.get('external'):
            assert 'ili' in elem
            elem.setdefault('meta')
        for defn in elem.get('definitions', []):
            defn.setdefault('text', '')
            defn.setdefault('meta')
        for rel in elem.get('relations', []):
            assert 'target' in rel
            assert 'relType' in rel
            rel.setdefault('meta')
        for ex in elem.get('examples', []):
            ex.setdefault('text', '')
            ex.setdefault('meta')
        if elem.get('lexicalized'):
            elem['lexicalized'] = False if elem['lexicalized'] == 'false' else True
        if elem.get('members'):
            elem['members'] = elem['members'].split()


def _validate_metadata(elem: _Elem) -> None:
    if elem.get('confidenceScore'):
        elem['confidenceScore'] = float(elem['confidenceScore'])


# Serialization ########################################################

def dump(resource: LexicalResource, destination: AnyPath) -> None:
    """Write wordnets in the WN-LMF format.

    Args:
        lexicons: a list of :class:`Lexicon` objects
    """
    version = resource['lmf_version']
    if version not in SUPPORTED_VERSIONS:
        raise LMFError(f'invalid version: {version}')
    destination = Path(destination).expanduser()
    doctype = _DOCTYPE.format(schema=_SCHEMAS[version])
    dc_uri = _DC_URIS[version]
    _version = version_info(version)
    with destination.open('wt', encoding='utf-8') as out:
        print(_XMLDECL.decode('utf-8'), file=out)
        print(doctype, file=out)
        print(f'<LexicalResource xmlns:dc="{dc_uri}">', file=out)
        for lexicon in resource['lexicons']:
            _dump_lexicon(lexicon, out, _version)
        print('</LexicalResource>', file=out)


def _dump_lexicon(
    lexicon: Union[Lexicon, LexiconExtension],
    out: TextIO,
    version: VersionInfo
) -> None:
    lexicontype = 'LexiconExtension' if lexicon.get('extends') else 'Lexicon'
    attrib = _build_lexicon_attrib(lexicon, version)
    attrdelim = '\n' + (' ' * len(f'  <{lexicontype} '))
    attrs = attrdelim.join(
        f'{attr}={quoteattr(str(val))}' for attr, val in attrib.items()
    )
    print(f'  <{lexicontype} {attrs}>', file=out)

    if version >= (1, 1):
        if lexicontype == 'LexiconExtension':
            assert lexicon.get('extends')
            lexicon = cast(LexiconExtension, lexicon)
            _dump_dependency(lexicon['extends'], 'Extends', out)
        for req in lexicon.get('requires', []):
            _dump_dependency(req, 'Requires', out)

    for entry in lexicon.get('entries', []):
        _dump_lexical_entry(entry, out, version)

    for synset in lexicon.get('synsets', []):
        _dump_synset(synset, out, version)

    if version >= (1, 1):
        for sb in lexicon.get('frames', []):
            _dump_syntactic_behaviour(sb, out, version)

    print(f'  </{lexicontype}>', file=out)


def _build_lexicon_attrib(
    lexicon: Union[Lexicon, LexiconExtension],
    version: VersionInfo
) -> Dict[str, str]:
    attrib = {
        'id': lexicon['id'],
        'label': lexicon['label'],
        'language': lexicon['language'],
        'email': lexicon['email'],
        'license': lexicon['license'],
        'version': lexicon['version'],
    }
    if lexicon.get('url'):
        attrib['url'] = lexicon['url']
    if lexicon.get('citation'):
        attrib['citation'] = lexicon['citation']
    if version >= (1, 1) and lexicon.get('logo'):
        attrib['logo'] = lexicon['logo']
    attrib.update(_meta_dict(lexicon.get('meta')))
    return attrib


def _dump_dependency(
    dep: Dependency, deptype: str, out: TextIO
) -> None:
    attrib = {'id': dep['id'], 'version': dep['version']}
    if dep.get('url'):
        attrib['url'] = dep['url']
    elem = ET.Element(deptype, attrib=attrib)
    print(_tostring(elem, 2), file=out)


def _dump_lexical_entry(
    entry: Union[LexicalEntry, ExternalLexicalEntry],
    out: TextIO,
    version: VersionInfo,
) -> None:
    frames = []
    attrib = {'id': entry['id']}
    if entry.get('external', False):
        elem = ET.Element('ExternalLexicalEntry', attrib=attrib)
        if entry.get('lemma'):
            assert entry['lemma'].get('external', False)
            elem.append(_build_lemma(entry['lemma'], version))
    else:
        entry = cast(LexicalEntry, entry)
        attrib.update(_meta_dict(entry.get('meta')))
        elem = ET.Element('LexicalEntry', attrib=attrib)
        elem.append(_build_lemma(entry['lemma'], version))
        if version < (1, 1):
            frames = [_build_syntactic_behaviour(sb, version)
                      for sb in entry.get('frames', [])]
    elem.extend([_build_form(form, version) for form in entry.get('forms', [])])
    elem.extend([_build_sense(sense, version)
                 for sense in entry.get('senses', [])])
    elem.extend(frames)
    print(_tostring(elem, 2), file=out)


def _build_lemma(
    lemma: Union[Lemma, ExternalLemma],
    version: VersionInfo
) -> ET.Element:
    if lemma.get('external', False):
        elem = ET.Element('ExternalLemma')
    else:
        lemma = cast(Lemma, lemma)
        attrib = {'writtenForm': lemma['writtenForm']}
        if lemma.get('script'):
            attrib['script'] = lemma['script']
        attrib['partOfSpeech'] = lemma['partOfSpeech']
        elem = ET.Element('Lemma', attrib=attrib)
    if version >= (1, 1):
        for pron in lemma.get('pronunciations', []):
            elem.append(_build_pronunciation(pron))
    for tag in lemma.get('tags', []):
        elem.append(_build_tag(tag))
    return elem


def _build_form(form: Union[Form, ExternalForm], version: VersionInfo) -> ET.Element:
    attrib = {}
    if version >= (1, 1) and form['id']:
        attrib['id'] = form['id']
    if form.get('external', False):
        elem = ET.Element('ExternalForm', attrib=attrib)
    else:
        form = cast(Form, form)
        attrib['writtenForm'] = form['writtenForm']
        if form.get('script'):
            attrib['script'] = form['script']
        elem = ET.Element('Form', attrib=attrib)
    if version >= (1, 1):
        for pron in form.get('pronunciations', []):
            elem.append(_build_pronunciation(pron))
    for tag in form.get('tags', []):
        elem.append(_build_tag(tag))
    return elem


def _build_pronunciation(pron: Pronunciation) -> ET.Element:
    attrib = {}
    if pron.get('variety'):
        attrib['variety'] = pron['variety']
    if pron.get('notation'):
        attrib['notation'] = pron['notation']
    if not pron.get('phonemic', True):
        attrib['phonemic'] = 'false'
    if pron.get('audio'):
        attrib['audio'] = pron['audio']
    elem = ET.Element('Pronunciation', attrib=attrib)
    elem.text = pron['text']
    return elem


def _build_tag(tag: Tag) -> ET.Element:
    elem = ET.Element('Tag', category=tag['category'])
    elem.text = tag['text']
    return elem


def _build_sense(
    sense: Union[Sense, ExternalSense],
    version: VersionInfo,
) -> ET.Element:
    attrib = {'id': sense['id']}
    if sense.get('external'):
        elem = ET.Element('ExternalSense', attrib=attrib)
    else:
        sense = cast(Sense, sense)
        attrib['synset'] = sense['synset']
        attrib.update(_meta_dict(sense.get('meta')))
        if not sense.get('lexicalized', True):
            attrib['lexicalized'] = 'false'
        if sense.get('adjposition'):
            attrib['adjposition'] = sense['adjposition']
        if version >= (1, 1) and sense.get('subcat'):
            attrib['subcat'] = ' '.join(sense['subcat'])
        elem = ET.Element('Sense', attrib=attrib)
    elem.extend([_build_relation(rel, 'SenseRelation')
                 for rel in sense.get('relations', [])])
    elem.extend([_build_example(ex) for ex in sense.get('examples', [])])
    elem.extend([_build_count(cnt) for cnt in sense.get('counts', [])])
    return elem


def _build_example(example: Example) -> ET.Element:
    elem = ET.Element('Example')
    elem.text = example['text']
    if example.get('language'):
        elem.set('language', example['language'])
    return elem


def _build_count(count: Count) -> ET.Element:
    elem = ET.Element('Count', attrib=_meta_dict(count.get('meta')))
    elem.text = str(count['value'])
    return elem


def _dump_synset(
    synset: Union[Synset, ExternalSynset],
    out: TextIO,
    version: VersionInfo
) -> None:
    attrib: Dict[str, str] = {'id': synset['id']}
    if synset.get('external', False):
        elem = ET.Element('ExternalSynset', attrib=attrib)
        elem.extend([_build_definition(defn) for defn in synset.get('definitions', [])])
    else:
        synset = cast(Synset, synset)
        attrib['ili'] = synset['ili']
        if synset.get('partOfSpeech'):
            attrib['partOfSpeech'] = synset['partOfSpeech']
        if not synset.get('lexicalized', True):
            attrib['lexicalized'] = 'false'
        if version >= (1, 1):
            if synset.get('members'):
                attrib['members'] = ' '.join(synset['members'])
            if synset.get('lexfile'):
                attrib['lexfile'] = synset['lexfile']
        attrib.update(_meta_dict(synset.get('meta')))
        elem = ET.Element('Synset', attrib=attrib)
        elem.extend([_build_definition(defn) for defn in synset.get('definitions', [])])
        if synset.get('ili_definition'):
            elem.append(_build_ili_definition(synset['ili_definition']))
    elem.extend([_build_relation(rel, 'SynsetRelation')
                 for rel in synset.get('relations', [])])
    elem.extend([_build_example(ex) for ex in synset.get('examples', [])])
    print(_tostring(elem, 2), file=out)


def _build_definition(definition: Definition) -> ET.Element:
    attrib = {}
    if definition.get('language'):
        attrib['language'] = definition['language']
    if definition.get('sourceSense'):
        attrib['sourceSense'] = definition['sourceSense']
    attrib.update(_meta_dict(definition.get('meta')))
    elem = ET.Element('Definition', attrib=attrib)
    elem.text = definition['text']
    return elem


def _build_ili_definition(ili_definition: ILIDefinition) -> ET.Element:
    elem = ET.Element('ILIDefinition', attrib=_meta_dict(ili_definition.get('meta')))
    elem.text = ili_definition['text']
    return elem


def _build_relation(relation: Relation, elemtype: str) -> ET.Element:
    attrib = {'target': relation['target'], 'relType': relation['relType']}
    attrib.update(_meta_dict(relation.get('meta')))
    return ET.Element(elemtype, attrib=attrib)


def _dump_syntactic_behaviour(
    syntactic_behaviour: SyntacticBehaviour,
    out: TextIO,
    version: VersionInfo
) -> None:
    elem = _build_syntactic_behaviour(syntactic_behaviour, version)
    print('    ' + _tostring(elem, 2), file=out)


def _build_syntactic_behaviour(
    syntactic_behaviour: SyntacticBehaviour,
    version: VersionInfo
) -> ET.Element:
    attrib = {'subcategorizationFrame': syntactic_behaviour['subcategorizationFrame']}
    if version >= (1, 1) and syntactic_behaviour.get('id'):
        attrib['id'] = syntactic_behaviour['id']
    elif version < (1, 1) and syntactic_behaviour.get('senses'):
        attrib['senses'] = ' '.join(syntactic_behaviour['senses'])
    return ET.Element('SyntacticBehaviour', attrib=attrib)


def _tostring(
        elem: ET.Element, level: int, short_empty_elements: bool = True
) -> str:
    _indent(elem, level)
    return ('  ' * level) + ET.tostring(
        elem,
        encoding='unicode',
        short_empty_elements=short_empty_elements
    )


def _indent(elem: ET.Element, level: int) -> None:
    self_indent = '\n' + '  ' * level
    child_indent = self_indent + '  '
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = child_indent
        for child in elem[:-1]:
            _indent(child, level + 1)
            child.tail = child_indent
        _indent(elem[-1], level + 1)
        elem[-1].tail = self_indent


def _meta_dict(meta: Optional[Metadata]) -> Dict[str, str]:
    if meta is not None:
        # Literal keys are required for typing purposes, so first
        # construct the dict and then remove those that weren't specified.
        d = {
            'dc:contributor': meta.get('contributor', ''),
            'dc:coverage': meta.get('coverage', ''),
            'dc:creator': meta.get('creator', ''),
            'dc:date': meta.get('date', ''),
            'dc:description': meta.get('description', ''),
            'dc:format': meta.get('format', ''),
            'dc:identifier': meta.get('identifier', ''),
            'dc:publisher': meta.get('publisher', ''),
            'dc:relation': meta.get('relation', ''),
            'dc:rights': meta.get('rights', ''),
            'dc:source': meta.get('source', ''),
            'dc:subject': meta.get('subject', ''),
            'dc:title': meta.get('title', ''),
            'dc:type': meta.get('type', ''),
            'status': meta.get('status', ''),
            'note': meta.get('note', ''),
        }
        d = {key: val for key, val in d.items() if val}
        # this one requires a conversion, so do it separately
        if 'confidenceScore' in meta:
            d['confidenceScore'] = str(meta['confidenceScore'])
    else:
        d = {}
    return d




"""
Wordnet Interface.
"""

__all__ = (
    '__version__',
    'Wordnet',
    'download',
    'add',
    'remove',
    'export',
    'projects',
    'lexicons',
    'Lexicon',
    'word',
    'words',
    'Word',
    'Form',
    'Pronunciation',
    'Tag',
    'sense',
    'senses',
    'Sense',
    'Count',
    'synset',
    'synsets',
    'Synset',
    'ili',
    'ilis',
    'ILI',
    'Error',
    'DatabaseError',
    'ConfigurationError',
    'ProjectError',
    'WnWarning',
)

from wn._exceptions import (
    Error,
    DatabaseError,
    ConfigurationError,
    ProjectError,
    WnWarning,
)
from wn._config import config  # noqa: F401
from wn._add import add, remove
from wn._export import export
from wn._download import download
from wn._core import (
    projects,
    lexicons, Lexicon,
    word, words, Word, Form, Pronunciation, Tag,
    sense, senses, Sense, Count,
    synset, synsets, Synset,
    ili, ilis, ILI,
    Wordnet
)

__version__ = '0.9.0'




from wn import util


def test_synset_id_formatter():
    f = util.synset_id_formatter
    assert f()(prefix='xyz', offset=123, pos='n') == 'xyz-00000123-n'
    assert f(prefix='xyz')(offset=123, pos='n') == 'xyz-00000123-n'
    assert f(prefix='xyz', pos='n')(offset=123) == 'xyz-00000123-n'
    assert f('abc-{offset}-{pos}')(offset=1, pos='v') == 'abc-1-v'




import pytest

import wn


@pytest.mark.usefixtures('mini_db')
def test_word_derived_words():
    assert len(wn.word('test-en-example-n').derived_words()) == 1
    assert len(wn.word('test-es-ejemplo-n').derived_words()) == 1


@pytest.mark.usefixtures('mini_db')
def test_synset_hypernyms():
    assert wn.synset('test-en-0002-n').hypernyms() == [
        wn.synset('test-en-0001-n')
    ]
    assert wn.synset('test-en-0001-n').hypernyms() == []


@pytest.mark.usefixtures('mini_db')
def test_synset_hypernyms_expand_default():
    assert wn.synset('test-es-0002-n').hypernyms() == [
        wn.synset('test-es-0001-n')
    ]
    assert wn.synset('test-es-0001-n').hypernyms() == []


@pytest.mark.usefixtures('mini_db')
def test_synset_hypernyms_expand_empty():
    w = wn.Wordnet(lang='es', expand='')
    assert w.synset('test-es-0002-n').hypernyms() == []


@pytest.mark.usefixtures('mini_db')
def test_synset_hypernyms_expand_specified():
    w = wn.Wordnet(lang='es', expand='test-en')
    assert w.synset('test-es-0002-n').hypernyms() == [
        w.synset('test-es-0001-n')
    ]


@pytest.mark.usefixtures('mini_db')
def test_synset_relations():
    w = wn.Wordnet(lang='en')
    assert w.synset('test-en-0002-n').relations() == {
        'hypernym': [w.synset('test-en-0001-n')],
        'hyponym': [w.synset('test-en-0004-n')]
    }


@pytest.mark.usefixtures('mini_db')
def test_sense_get_related():
    w = wn.Wordnet('test-en')
    assert w.sense('test-en-example-n-0002-01').get_related() == [
        w.sense('test-en-exemplify-v-0003-01')
    ]


@pytest.mark.usefixtures('mini_db')
def test_sense_relations():
    w = wn.Wordnet('test-en')
    assert w.sense('test-en-example-n-0002-01').relations() == {
        'derivation': [w.sense('test-en-exemplify-v-0003-01')]
    }


@pytest.mark.usefixtures('mini_db_1_1')
def test_extension_relations():
    # default mode
    assert wn.synset('test-en-0007-v').hypernyms() == [
        wn.synset('test-en-ext-0009-v')
    ]
    assert wn.synset('test-en-ext-0009-v').hyponyms() == [
        wn.synset('test-en-0007-v')
    ]
    assert wn.sense('test-en-information-n-0001-01').get_related('pertainym') == [
        wn.sense('test-en-ext-info-n-0001-01')
    ]
    assert wn.sense('test-en-ext-info-n-0001-01').get_related('pertainym') == [
        wn.sense('test-en-information-n-0001-01')
    ]

    # restricted to base
    w = wn.Wordnet(lexicon='test-en')
    assert w.synset('test-en-0007-v').hypernyms() == []
    assert w.sense('test-en-information-n-0001-01').get_related('pertainym') == []

    # base and extension
    w = wn.Wordnet(lexicon='test-en test-en-ext')
    assert w.synset('test-en-0007-v').hypernyms() == [
        w.synset('test-en-ext-0009-v')
    ]
    assert w.synset('test-en-ext-0009-v').hyponyms() == [
        w.synset('test-en-0007-v')
    ]
    assert w.sense('test-en-information-n-0001-01').get_related('pertainym') == [
        w.sense('test-en-ext-info-n-0001-01')
    ]
    assert w.sense('test-en-ext-info-n-0001-01').get_related('pertainym') == [
        w.sense('test-en-information-n-0001-01')
    ]

    # restricted to extension
    w = wn.Wordnet(lexicon='test-en-ext')
    assert w.synset('test-en-ext-0009-v').hyponyms() == []
    assert w.sense('test-en-ext-info-n-0001-01').get_related('pertainym') == []




"""Web interface for Wn databases."""

from typing import Optional, Union
from functools import wraps
from urllib.parse import urlsplit, parse_qs, urlencode

from starlette.applications import Starlette  # type: ignore
from starlette.responses import JSONResponse  # type: ignore
from starlette.routing import Route  # type: ignore
from starlette.requests import Request  # type: ignore

import wn

DEFAULT_PAGINATION_LIMIT = 50


def paginate(proto):

    def paginate_wrapper(func):

        @wraps(func)
        async def _paginate_wrapper(request: Request) -> JSONResponse:
            url = str(request.url)
            query = dict(request.query_params)
            offset = abs(int(query.pop('page[offset]', 0)))
            limit = abs(int(query.pop('page[limit]', DEFAULT_PAGINATION_LIMIT)))

            obj = await func(request)
            total = len(obj['data'])
            prev = max(0, offset - limit)
            next = offset + limit
            last = (total//limit)*limit

            obj['data'] = [proto(x, request) for x in obj['data'][offset:offset+limit]]
            obj.setdefault('meta', {}).update(total=total)

            links = {}
            if offset > 0:
                links['first'] = replace_query_params(url, **{'page[offset]': 0})
                links['prev'] = replace_query_params(url, **{'page[offset]': prev})
            if next < total:
                links['next'] = replace_query_params(url, **{'page[offset]': next})
                links['last'] = replace_query_params(url, **{'page[offset]': last})
            if links:
                obj.setdefault('links', {}).update(links)

            return JSONResponse(obj)

        return _paginate_wrapper

    return paginate_wrapper


def replace_query_params(url: str, **params) -> str:
    u = urlsplit(url)
    q = parse_qs(u.query)
    q.update(params)
    qs = urlencode(q, doseq=True)
    return u._replace(query=qs).geturl()


# Data-making functions

def _url_for_obj(
    request: Request,
    name: str,
    obj: Union[wn.Word, wn.Sense, wn.Synset],
    lexicon: Optional[str] = None,
) -> str:
    if lexicon is None:
        lexicon = obj.lexicon().specifier()
    kwargs = {
        'lexicon': lexicon,
        name: obj.id
    }
    return request.url_for(name, **kwargs)


def make_lexicon(lex: wn.Lexicon, request: Request) -> dict:
    spec = lex.specifier()
    return {
        'id': spec,
        'type': 'lexicon',
        'attributes': {
            # cannot have 'id' as JSON:API disallows it
            'version': lex.version,
            'label': lex.label,
            'language': lex.language,
            'license': lex.license,
        },
        'links': {
            'self': request.url_for('lexicon', lexicon=spec)
        },
        'relationships': {
            'words': {
                'links': {'related': request.url_for('words', lexicon=spec)},
            },
            'synsets': {
                'links': {'related': request.url_for('synsets', lexicon=spec)},
            },
            'senses': {
                'links': {'related': request.url_for('senses', lexicon=spec)},
            },
        }
    }


def make_word(w: wn.Word, request: Request, basic: bool = False) -> dict:
    lex_spec = w.lexicon().specifier()
    d: dict = {
        'id': w.id,
        'type': 'word',
        'attributes': {
            'pos': w.pos,
            'lemma': w.lemma(),
            'forms': w.forms(),
        },
        'links': {
            'self': _url_for_obj(request, 'word', w, lexicon=lex_spec)
        }
    }
    if not basic:
        synsets = w.synsets()
        lex_link = request.url_for('lexicon', lexicon=lex_spec)
        senses_link = request.url_for('senses', word=w.id, lexicon=lex_spec)
        d.update({
            'relationships': {
                'senses': {'links': {'related': senses_link}},
                'synsets': {'data': [dict(type='synset', id=ss.id) for ss in synsets]},
                'lexicon': {'links': {'related': lex_link}}
            },
            'included': [make_synset(ss, request, basic=True) for ss in synsets]
        })
    return d


def make_sense(s: wn.Sense, request: Request, basic: bool = False) -> dict:
    lex_spec = s.lexicon().specifier()
    d: dict = {
        'id': s.id,
        'type': 'sense',
        'links': {
            'self': _url_for_obj(request, 'sense', s, lexicon=lex_spec)
        }
    }
    if not basic:
        w = s.word()
        ss = s.synset()
        lex_link = request.url_for('lexicon', lexicon=lex_spec)
        word_link = request.url_for('word', word=w.id, lexicon=lex_spec)
        synset_link = request.url_for('synset', synset=ss.id, lexicon=lex_spec)
        relationships: dict = {
            'word': {'links': {'related': word_link}},
            'synset': {'links': {'related': synset_link}},
            'lexicon': {'links': {'related': lex_link}}
        }
        included = []
        for relname, slist in s.relations().items():
            relationships[relname] = {
                'data': [dict(type='sense', id=_s.id) for _s in slist]
            }
            included.extend([make_sense(_s, request, basic=True) for _s in slist])
        d.update({'relationships': relationships, 'included': included})
    return d


def make_synset(ss: wn.Synset, request: Request, basic: bool = False) -> dict:
    lex_spec = ss.lexicon().specifier()
    d: dict = {
        'id': ss.id,
        'type': 'synset',
        'attributes': {
            'pos': ss.pos,
            'ili': ss._ili,
        },
        'links': {
            'self': _url_for_obj(request, 'synset', ss, lexicon=lex_spec)
        }
    }
    if not basic:
        words = ss.words()
        lex_link = request.url_for('lexicon', lexicon=lex_spec)
        members_link = request.url_for('senses', synset=ss.id, lexicon=lex_spec)
        relationships: dict = {
            'members': {'links': {'related': members_link}},
            'words': {'data': [dict(type='word', id=w.id) for w in words]},
            'lexicon': {'links': {'related': lex_link}}
        }
        included = [make_word(w, request, basic=True) for w in words]
        for relname, sslist in ss.relations().items():
            relationships[relname] = {
                'data': [dict(type='synset', id=_s.id) for _s in sslist]
            }
            included.extend([make_synset(_s, request, basic=True) for _s in sslist])
        d.update({'relationships': relationships, 'included': included})
    return d


@paginate(make_lexicon)
async def lexicons(request):
    query = request.query_params
    _lexicons = wn.lexicons(
        lexicon=query.get('lexicon'),
        lang=query.get('lang'),
    )
    return {'data': _lexicons}


async def lexicon(request):
    path_params = request.path_params
    lex = wn.lexicons(lexicon=path_params['lexicon'])[0]
    return JSONResponse({'data': make_lexicon(lex, request)})


@paginate(make_word)
async def words(request):
    query = request.query_params
    lexicon = request.path_params.get('lexicon') or query.get('lexicon')
    _words = wn.words(
        form=query.get('form'),
        pos=query.get('pos'),
        lexicon=lexicon,
        lang=query.get('lang'),
    )
    return {'data': _words}


async def word(request):
    path_params = request.path_params
    word = wn.word(path_params['word'], lexicon=path_params['lexicon'])
    return JSONResponse({'data': make_word(word, request)})


@paginate(make_sense)
async def senses(request):
    query = request.query_params
    path = request.path_params
    lexicon = path.get('lexicon') or query.get('lexicon')
    if 'word' in path:
        _senses = wn.word(path['word'], lexicon=lexicon).senses()
    elif 'synset' in path:
        _senses = wn.synset(path['synset'], lexicon=lexicon).members()
    else:
        _senses = wn.senses(
            form=query.get('form'),
            pos=query.get('pos'),
            lexicon=lexicon,
            lang=query.get('lang'),
        )
    return {'data': _senses}


async def sense(request):
    path_params = request.path_params
    sense = wn.sense(path_params['sense'], lexicon=path_params['lexicon'])
    return JSONResponse({'data': make_sense(sense, request)})


@paginate(make_synset)
async def synsets(request):
    query = request.query_params
    lexicon = request.path_params.get('lexicon') or query.get('lexicon')
    _synsets = wn.synsets(
        form=query.get('form'),
        pos=query.get('pos'),
        ili=query.get('ili'),
        lexicon=lexicon,
        lang=query.get('lang'),
    )
    return {'data': _synsets}


async def synset(request):
    path_params = request.path_params
    synset = wn.synset(path_params['synset'], lexicon=path_params['lexicon'])
    return JSONResponse({'data': make_synset(synset, request)})


routes = [
    Route('/lexicons', endpoint=lexicons),
    Route('/lexicons/{lexicon}', endpoint=lexicon),
    Route('/lexicons/{lexicon}/words', endpoint=words),
    Route('/lexicons/{lexicon}/words/{word}', endpoint=word),
    Route('/lexicons/{lexicon}/words/{word}/senses', endpoint=senses),
    Route('/lexicons/{lexicon}/senses', endpoint=senses),
    Route('/lexicons/{lexicon}/senses/{sense}', endpoint=sense),
    Route('/lexicons/{lexicon}/synsets', endpoint=synsets),
    Route('/lexicons/{lexicon}/synsets/{synset}', endpoint=synset),
    Route('/lexicons/{lexicon}/synsets/{synset}/members', endpoint=senses),
    Route('/words', endpoint=words),
    Route('/senses', endpoint=senses),
    Route('/synsets', endpoint=synsets),
]

app = Starlette(debug=True, routes=routes)




from typing import List, Dict, Set, Tuple, Sequence, Optional, cast

import wn
from wn._types import AnyPath, VersionInfo
from wn._util import version_info
from wn import lmf
from wn._queries import (
    find_entries,
    find_senses,
    find_synsets,
    find_syntactic_behaviours,
    find_proposed_ilis,
    get_entry_senses,
    get_sense_relations,
    get_sense_synset_relations,
    get_synset_relations,
    get_synset_members,
    get_examples,
    get_definitions,
    get_metadata,
    get_lexicalized,
    get_adjposition,
    get_form_pronunciations,
    get_form_tags,
    get_sense_counts,
    get_lexfile,
    get_lexicon,
    get_lexicon_dependencies,
    get_lexicon_extension_bases,
)
from wn._core import Lexicon


def export(
        lexicons: Sequence[Lexicon], destination: AnyPath, version: str = '1.0'
) -> None:
    """Export lexicons from the database to a WN-LMF file.

    More than one lexicon may be exported in the same file, subject to
    these conditions:

    - identifiers on wordnet entities must be unique in all lexicons
    - lexicons extensions may not be exported with their dependents

    >>> w = wn.Wordnet(lexicon='cmnwn zsmwn')
    >>> wn.export(w.lexicons(), 'cmn-zsm.xml')

    Args:
        lexicons: sequence of :class:`wn.Lexicon` objects
        destination: path to the destination file
        version: LMF version string

    """
    _precheck(lexicons)
    assert version in lmf.SUPPORTED_VERSIONS
    _version = version_info(version)
    resource: lmf.LexicalResource = {
        'lmf_version': version,
        'lexicons': [_export_lexicon(lex, _version) for lex in lexicons]
    }
    lmf.dump(resource, destination)


def _precheck(lexicons: Sequence[Lexicon]) -> None:
    all_ids: Set[str] = set()
    for lex in lexicons:
        lexids = (lex._id,)
        idset = {lex.id}
        idset.update(row[0] for row in find_entries(lexicon_rowids=lexids))
        idset.update(row[0] for row in find_senses(lexicon_rowids=lexids))
        idset.update(row[0] for row in find_synsets(lexicon_rowids=lexids))
        # TODO: syntactic behaviours
        if all_ids.intersection(idset):
            raise wn.Error('cannot export: non-unique identifiers in lexicons')
        all_ids |= idset


_SBMap = Dict[str, List[Tuple[str, str]]]


def _export_lexicon(lexicon: Lexicon, version: VersionInfo) -> lmf.Lexicon:
    lexids = (lexicon._id,)

    # WN-LMF 1.0 lexicons put syntactic behaviours on lexical entries
    # WN-LMF 1.1 lexicons use a 'subcat' IDREFS attribute
    sbmap: _SBMap = {}
    if version < (1, 1):
        for sbid, frame, sids in find_syntactic_behaviours(lexicon_rowids=lexids):
            for sid in sids:
                sbmap.setdefault(sid, []).append((sbid, frame))

    lex: lmf.Lexicon = {
        'id': lexicon.id,
        'label': lexicon.label,
        'language': lexicon.language,
        'email': lexicon.email,
        'license': lexicon.license,
        'version': lexicon.version,
        'url': lexicon.url or '',
        'citation': lexicon.citation or '',
        'entries': _export_lexical_entries(lexids, sbmap, version),
        'synsets': _export_synsets(lexids, version),
        'meta': _export_metadata(lexicon._id, 'lexicons'),
    }
    if version >= (1, 1):
        lex['logo'] = lexicon.logo or ''
        lex['requires'] = _export_requires(lexicon._id)
        lex['frames'] = _export_syntactic_behaviours_1_1(lexids)

    return lex


def _export_requires(lexid: int) -> List[lmf.Dependency]:
    return [
        {'id': id, 'version': version, 'url': url}
        for id, version, url, _ in get_lexicon_dependencies(lexid)
    ]


def _export_extends(lexid: int) -> lmf.Dependency:
    ext_lexid = get_lexicon_extension_bases(lexid, depth=1)[0]
    _, id, _, _, _, _, version, url, *_ = get_lexicon(ext_lexid)
    return {'id': id, 'version': version, 'url': url}


def _export_lexical_entries(
    lexids: Sequence[int],
    sbmap: _SBMap,
    version: VersionInfo
) -> List[lmf.LexicalEntry]:
    entries: List[lmf.LexicalEntry] = []
    for id, pos, forms, _, rowid in find_entries(lexicon_rowids=lexids):
        entry: lmf.LexicalEntry = {
            'id': id,
            'lemma': {
                'writtenForm': forms[0][0],
                'partOfSpeech': pos,
                'script': forms[0][2] or '',
                'tags': _export_tags(forms[0][3]),
            },
            'forms': [],
            'senses': _export_senses(rowid, lexids, sbmap, version),
            'meta': _export_metadata(rowid, 'entries'),
        }
        if version >= (1, 1):
            entry['lemma']['pronunciations'] = _export_pronunciations(forms[0][3])
        for form, fid, script, frowid in forms[1:]:
            _form: lmf.Form = {
                'id': fid or '',
                'writtenForm': form,
                'script': script or '',
                'tags': _export_tags(frowid),
            }
            if version >= (1, 1):
                _form['pronunciations'] = _export_pronunciations(frowid)
            entry['forms'].append(_form)
        if version < (1, 1):
            entry['frames'] = _export_syntactic_behaviours_1_0(entry, sbmap)
        entries.append(entry)
    return entries


def _export_pronunciations(rowid: int) -> List[lmf.Pronunciation]:
    return [
        {'text': text,
         'variety': variety,
         'notation': notation,
         'phonemic': phonemic,
         'audio': audio}
        for text, variety, notation, phonemic, audio
        in get_form_pronunciations(rowid)
    ]


def _export_tags(rowid: int) -> List[lmf.Tag]:
    return [
        {'text': text, 'category': category}
        for text, category in get_form_tags(rowid)
    ]


def _export_senses(
    entry_rowid: int,
    lexids: Sequence[int],
    sbmap: _SBMap,
    version: VersionInfo,
) -> List[lmf.Sense]:
    senses: List[lmf.Sense] = []
    for id, _, synset, _, rowid in get_entry_senses(entry_rowid, lexids):
        sense: lmf.Sense = {
            'id': id,
            'synset': synset,
            'relations': _export_sense_relations(rowid, lexids),
            'examples': _export_examples(rowid, 'senses', lexids),
            'counts': _export_counts(rowid, lexids),
            'lexicalized': get_lexicalized(rowid, 'senses'),
            'adjposition': get_adjposition(rowid) or '',
            'meta': _export_metadata(rowid, 'senses'),
        }
        if version >= (1, 1) and id in sbmap:
            sense['subcat'] = sorted(sbid for sbid, _ in sbmap[id])
        senses.append(sense)
    return senses


def _export_sense_relations(
    sense_rowid: int,
    lexids: Sequence[int]
) -> List[lmf.Relation]:
    relations: List[lmf.Relation] = [
        {'target': id,
         'relType': type,
         'meta': _export_metadata(rowid, 'sense_relations')}
        for type, rowid, id, *_
        in get_sense_relations(sense_rowid, '*', lexids)
    ]
    relations.extend(
        {'target': id,
         'relType': type,
         'meta': _export_metadata(rowid, 'sense_synset_relations')}
        for type, rowid, id, *_
        in get_sense_synset_relations(sense_rowid, '*', lexids)
    )
    return relations


def _export_examples(
    rowid: int,
    table: str,
    lexids: Sequence[int]
) -> List[lmf.Example]:
    return [
        {'text': text,
         'language': language,
         'meta': _export_metadata(rowid, f'{table[:-1]}_examples')}
        for text, language, rowid
        in get_examples(rowid, table, lexids)
    ]


def _export_counts(rowid: int, lexids: Sequence[int]) -> List[lmf.Count]:
    return [
        {'value': val,
         'meta': _export_metadata(id, 'counts')}
        for val, id in get_sense_counts(rowid, lexids)
    ]


def _export_synsets(lexids: Sequence[int], version: VersionInfo) -> List[lmf.Synset]:
    synsets: List[lmf.Synset] = []
    for id, pos, ili, _, rowid in find_synsets(lexicon_rowids=lexids):
        ilidef = _export_ili_definition(rowid)
        if ilidef and not ili:
            ili = 'in'  # special case for proposed ILIs
        ss: lmf.Synset = {
            'id': id,
            'ili': ili or '',
            'partOfSpeech': pos,
            'definitions': _export_definitions(rowid, lexids),
            'relations': _export_synset_relations(rowid, lexids),
            'examples': _export_examples(rowid, 'synsets', lexids),
            'lexicalized': get_lexicalized(rowid, 'synsets'),
            'lexfile': get_lexfile(rowid) or '',
            'meta': _export_metadata(rowid, 'synsets'),
        }
        if ilidef:
            ss['ili_definition'] = ilidef
        if version >= (1, 1):
            ss['members'] = [row[0] for row in get_synset_members(rowid, lexids)]
        synsets.append(ss)
    return synsets


def _export_definitions(rowid: int, lexids: Sequence[int]) -> List[lmf.Definition]:
    return [
        {'text': text,
         'language': language,
         'sourceSense': sense_id,
         'meta': _export_metadata(rowid, 'definitions')}
        for text, language, sense_id, rowid
        in get_definitions(rowid, lexids)
    ]


def _export_ili_definition(synset_rowid: int) -> Optional[lmf.ILIDefinition]:
    _, _, defn, rowid = next(find_proposed_ilis(synset_rowid=synset_rowid),
                             (None, None, None, None))
    ilidef: Optional[lmf.ILIDefinition] = None
    if defn:
        meta = None
        if rowid is not None:
            meta = _export_metadata(rowid, 'proposed_ilis')
        ilidef = {'text': defn, 'meta': meta}
    return ilidef


def _export_synset_relations(
    synset_rowid: int,
    lexids: Sequence[int]
) -> List[lmf.Relation]:
    return [
        {'target': id,
         'relType': type,
         'meta': _export_metadata(rowid, 'synset_relations')}
        for type, rowid, id, *_
        in get_synset_relations((synset_rowid,), '*', lexids)
    ]


def _export_syntactic_behaviours_1_0(
    entry: lmf.LexicalEntry,
    sbmap: _SBMap,
) -> List[lmf.SyntacticBehaviour]:
    frames: List[lmf.SyntacticBehaviour] = []
    sense_ids = {s['id'] for s in entry.get('senses', [])}
    sbs: Dict[str, Set[str]] = {}
    for sid in sense_ids:
        for _, subcat_frame in sbmap.get(sid, []):
            sbs.setdefault(subcat_frame, set()).add(sid)
    for subcat_frame, sids in sbs.items():
        frame: lmf.SyntacticBehaviour = {
            'subcategorizationFrame': subcat_frame,
            'senses': sorted(sids),
        }
        frames.append(frame)
    return frames


def _export_syntactic_behaviours_1_1(
    lexids: Sequence[int]
) -> List[lmf.SyntacticBehaviour]:
    return [
        {'id': id or '',
         'subcategorizationFrame': frame}
        for id, frame, _ in find_syntactic_behaviours(lexicon_rowids=lexids)
    ]


def _export_metadata(rowid: int, table: str) -> lmf.Metadata:
    return cast(lmf.Metadata, get_metadata(rowid, table))
